# SPDX-FileCopyrightText: 2017-2023 Contributors to the OpenSTEF project <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0

"""This module contains all wheather related functions used for feature engineering."""
from typing import Union

import numpy as np
import pandas as pd
import pvlib
import structlog
from pvlib.location import Location

from openstef.data_classes.prediction_job import PredictionJobDataClass

logger = structlog.get_logger(__name__)


# Set some (nameless) constants for the Antoine equation:
A: float = 6.116
M: float = 7.6
TN: float = 240.7
# Set some constants
TORR: float = 133.322368  # 1 torr = 133 Pa
# 1.168 is the mass of 1 m^3 of air on sea level with standard pressure.
D: float = 1.168

DEFAULT_LAT: float = 52.132633
DEFAULT_LON: float = 5.291266


def calc_saturation_pressure(
    temperature: Union[float, np.ndarray]
) -> Union[float, np.ndarray]:
    """Calculate the water vapour pressure from the temperature.

    See https://www.vaisala.com/sites/default/files/documents/Humidity_Conversion_Formulas_B210973EN-F.pdf.

    Args:
        temperature: Temperature in C
    Returns:
        The saturation pressure of water at the respective temperature

    """
    psat = A * 10 ** ((M * temperature) / (temperature + TN))
    return psat


def calc_vapour_pressure(
    rh: Union[float, np.ndarray], psat: Union[float, np.ndarray]
) -> Union[float, np.ndarray]:
    """Calculates the vapour pressure.

    Args:
        rh: Relative humidity
        psat: Saturation pressure: see calc_saturation_pressure
    Returns:
        The water vapour pressure

    """
    return rh * psat


def calc_dewpoint(
    vapour_pressure: Union[float, np.ndarray]
) -> Union[float, np.ndarray]:
    """Calculates the dewpoint, see https://en.wikipedia.org/wiki/Dew_point for mroe info.

    Args:
        vapour_pressure: The vapour pressure for which the dewpoint should be calculated
    Returns:
        Dewpoint

    """
    return TN / ((M / np.log10(vapour_pressure / A)) - 1)


def calc_air_density(
    temperature: Union[float, np.ndarray],
    pressure: Union[float, np.ndarray],
    rh: Union[float, np.ndarray],
) -> Union[float, np.ndarray]:
    """Calculates the dewpoint.

    Args:
        temperature: The temperature in C
        pressure: the atmospheric pressure in Pa
        rh: Relative humidity

    Returns:
        The air density (kg/m^3)

    """
    # Calculate saturation pressure
    psat = calc_saturation_pressure(temperature)
    # Calculate the current vapour pressure
    vapour_pressure = calc_vapour_pressure(rh, psat)

    # Set tempareture to K
    temperature_k = temperature + 273.15

    # Calculate air density
    air_density = (
        D
        * (273.15 / temperature_k)
        * ((pressure - 0.3783 * vapour_pressure) / 760 / TORR)
    )

    return air_density


def add_humidity_features(
    data: pd.DataFrame, feature_names: list[str] = None
) -> pd.DataFrame:
    """Adds humidity features to the input dataframe.

    These features are calculated using functions defines in this module. A list of
    requested features is used to determine whether to add the humidity features or not.

    Args:
        data: Input dataframe to which features have to be added
        feature_names: list of requested features.

    Returns:
        Same as input dataframe with extra columns for the humidty features.

    """
    # If features is none add humidity feature anyway
    if feature_names is None:
        humidity_features = True

    # Otherwise check if they are among the reuqested features
    else:
        humidity_features = any(
            x
            in [
                "saturation_pressure",
                "vapour_pressure",
                "dewpoint",
                "air_density",
            ]
            for x in feature_names
        )

    # Check if any of the humidity features are requested and add them
    if humidity_features:
        # Try to add humidity  calculations, ignore if required columns are missing
        try:
            humidity_df = humidity_calculations(data.temp, data.humidity, data.pressure)
            data = data.join(humidity_df)
        except AttributeError:
            pass  # This happens when a required column for humidity_calculations
            # is not present

    return data


def humidity_calculations(
    temperature: Union[float, np.ndarray],
    rh: Union[float, np.ndarray],
    pressure: Union[float, np.ndarray],
) -> Union[dict, np.ndarray]:
    """Function that calculates weather features based on humidity..

    These features are:
        - Saturation pressure
        - Vapour pressure
        - Dewpoint
        - Air density

    Args:
        temperature: Temperature in C
        rh: Relative humidity in %
        pressure: The air pressure in hPa

    Returns:
        If the input is an np.ndarray; a pandas dataframe with the calculated moisture indices,
        if the input is numeric; a dict with the calculated moisture indices

    """
    # First: a sanity check on the relative humidity and the air pressure
    # We only check on the type of temperature, because they need to be the same anyway
    is_series = isinstance(temperature, (np.ndarray, pd.Series))
    is_scalar = isinstance(temperature, (float, int))

    if is_scalar is False and is_series is False:
        raise TypeError(
            "The input should be a pandas series or np.ndarry, or float or int"
        )

    # Suppres copy warnings
    with pd.option_context("mode.chained_assignment", None):
        if is_series:
            rh[rh > 1] = rh / 100  # This triggers copy warnings
            pressure[pressure < 80000] = np.nan  # This triggers copy warnings
        else:
            if rh > 1:
                rh /= 100
            if pressure < 80000:
                pressure = np.nan

    # If the input is a dataframe or np.ndarrays: return a dataframe
    if is_series:
        humidity_df = pd.DataFrame(
            columns=[
                "saturation_pressure",
                "vapour_pressure",
                "dewpoint",
                "air_density",
            ]
        )
        humidity_df["saturation_pressure"] = calc_saturation_pressure(temperature)
        humidity_df["vapour_pressure"] = calc_vapour_pressure(
            rh, humidity_df.saturation_pressure
        )
        humidity_df["dewpoint"] = calc_dewpoint(humidity_df.vapour_pressure)
        humidity_df["air_density"] = calc_air_density(temperature, pressure, rh)

        return humidity_df

    # Else: if the input is numeric: return a dict
    psat = calc_saturation_pressure(temperature)
    pw = calc_vapour_pressure(rh, psat)
    td = calc_dewpoint(pw)
    air_density = calc_air_density(temperature, pressure, rh)
    return {
        "saturation_pressure": psat,
        "vapour_pressure": pw,
        "dewpoint": td,
        "air_density": air_density,
    }


def calculate_windspeed_at_hubheight(
    windspeed: Union[float, pd.Series],
    fromheight: float = 10.0,
    hub_height: float = 100.0,
) -> pd.Series:
    """Calculate windspeed at hubheight.

    Calculates the windspeed at hubheigh by extrapolation from a given height to a given
    hub height using the wind power law https://en.wikipedia.org/wiki/Wind_profile_power_law

    Args:
        windspeed: float OR pandas series of windspeed at height = height
        fromheight: height (m) of the windspeed data. Default is 10m
        hubheight: height (m) of the turbine

    Returns:
        Windspeed at hubheight.

    """
    alpha = 0.143

    if not isinstance(windspeed, (np.ndarray, float, int, pd.Series)):
        raise TypeError(
            "The windspeed is not of the expected type!\n                        Got"
            " {}, expected np.ndarray, pd series or numeric".format(type(windspeed))
        )

    try:
        if any(windspeed < 0):
            raise ValueError(
                "The windspeed cannot be negative, as it is the lenght of a vector"
            )
    except TypeError:
        if windspeed < 0:
            raise ValueError(
                "The windspeed cannot be negative, as it is the lenght of a vector"
            )
        windspeed = abs(windspeed)

    return windspeed * (hub_height / fromheight) ** alpha


def calculate_windturbine_power_output(
    windspeed: pd.Series, n_turbines: int = 1, turbine_data: dict = None
) -> pd.Series:
    """Calculate wind turbine power output.

    These values are related through the power curve, which is described by turbine_data.
    If no turbine_data is given, default values are used and results are normalized to 1MWp.
    If n_turbines=0, the result is normalized to a rated power of 1.

    Args:
        windspeed: pd.DataFrame(index = datetime, columns = ["windspeedHub"])
        nTurbines: The number of turbines
        turbineData: slope_center, rated_power, steepness

    Returns:
        pd.DataFrame(index = datetime, columns = ["forecast"])

    """
    if turbine_data is None:
        turbine_data = {
            "name": "Lagerwey L100",  # not used here
            "cut_in": 3,  # not used here
            "cut_off": 25,  # not used here
            "kind": "onshore",  # not used here
            "manufacturer": "Lagerwey",  # not used here
            "peak_capacity": 1,  # not used here
            "rated_power": 1,
            "slope_center": 8.07,
            "steepness": 0.664,
        }
    else:
        required_properties = ["rated_power", "steepness", "slope_center"]
        for prop in required_properties:
            if prop not in turbine_data.keys():
                raise KeyError(f"Required property '{prop}' not set in turbine data")

    generated_power = turbine_data["rated_power"] / (
        1
        + np.exp(
            -turbine_data["steepness"] * (windspeed - turbine_data["slope_center"])
        )
    )
    generated_power *= n_turbines

    return generated_power


def add_additional_wind_features(
    data: pd.DataFrame, feature_names: list[str] = None
) -> pd.DataFrame:
    """Adds additional wind features to the input data.

    Args:
        data: Dataframe to which the wind features have to be added
        feature_names: List of requested features

    Returns:
        DataFrame same as input dataframe with extra columns for the added wind features

    """
    if feature_names is None:
        additional_wind_features = True
    else:
        additional_wind_features = any(
            x
            in [
                "windspeed_100mExtrapolated",
                "windPowerFit_extrapolated",
                "windpowerFit_harm_arome",
            ]
            for x in feature_names
        )

    # Add add_additional_wind_features
    if "windspeed" in data.columns and additional_wind_features:
        data["windspeed_100mExtrapolated"] = calculate_windspeed_at_hubheight(
            data["windspeed"]
        )

        data["windPowerFit_extrapolated"] = calculate_windturbine_power_output(
            data["windspeed_100mExtrapolated"]
        )

    # Do extra check
    if "windspeed_100m" in data.columns and additional_wind_features:
        data["windpowerFit_harm_arome"] = calculate_windturbine_power_output(
            data["windspeed_100m"].astype(float)
        )

    return data


def calculate_dni(radiation: pd.Series, pj: PredictionJobDataClass) -> pd.Series:
    """Calculate the direct normal irradiance (DNI).

    This function uses the predicted radiation and information derived from the location (obtained from pj)


    Args:
        radiation: predicted radiation including DatetimeIndex with right time-zone
        pj: PredictJob including information about the location (lat, lon)

    Returns:
        Direct normal irradiance (DNI).

    """
    loc = Location(pj.get("lat", DEFAULT_LAT), pj.get("lon", DEFAULT_LON), tz="utc")
    times = radiation.index

    # calculate data for loc(ation) at times with clear_sky, as if there would be a clear sky.
    cs = loc.get_clearsky(times)

    # get solar position variable(s) for loc(ation) at times
    solpos = pvlib.solarposition.get_solarposition(times, loc.latitude, loc.longitude)
    solar_zenith = solpos.apparent_zenith

    # convert radiation (ghi) to right unit (J/m^2 to kWh/m^2)
    # TODO: check whether unit conversion is necessary
    ghi_forecasted = radiation / 3600
    # convert ghi to dni
    dni_converted = pvlib.irradiance.dni(
        ghi_forecasted, cs.dhi, solar_zenith, clearsky_dni=cs.dni
    )
    dni_converted = dni_converted.fillna(0)
    return dni_converted


def calculate_gti(
    radiation: pd.Series,
    pj: PredictionJobDataClass,
    surface_tilt: float = 34.0,
    surface_azimuth: float = 180,
) -> pd.Series:
    """Calculate the GTI/POA using the radiation.

    This function assumes Global Tilted Irradiance (GTI) = Plane of Array (POA)

    Args:
        radiation: pandas series with DatetimeIndex with right timezone information
        pj: prediction job which should at least contain the latitude and longitude location.
        surface_tilt: The tilt of the surface of, for example, your PhotoVoltaic-system.
        surface_azimuth: The way the surface is facing. South facing 180 degrees, North facing 0 degrees, East facing 90 degrees and West facing 270 degrees

    Returns:
        Global Tilted Irradiance (GTI)

    """
    loc = Location(pj.get("lat", DEFAULT_LAT), pj.get("lon", DEFAULT_LON), tz="utc")
    times = radiation.index

    # calculate data for loc(ation) at times with clear_sky, as if there would be a clear sky.
    cs = loc.get_clearsky(times)
    dni = calculate_dni(radiation, pj)

    # get solar position variable(s) for loc(ation) at times
    solpos = pvlib.solarposition.get_solarposition(times, loc.latitude, loc.longitude)
    solar_zenith = solpos.apparent_zenith
    solar_azimuth = solpos.azimuth

    ghi_forecasted = radiation / 3600
    gti = pvlib.irradiance.get_total_irradiance(
        surface_tilt,
        surface_azimuth,
        solar_zenith,
        solar_azimuth,
        dni=dni,
        ghi=ghi_forecasted,
        dhi=cs.dhi,
    )

    return gti["poa_global"]


def add_additional_solar_features(
    data: pd.DataFrame,
    pj: PredictionJobDataClass = None,
    feature_names: list[str] = None,
) -> pd.DataFrame:
    """Adds additional solar features to the input data.

    Args:
        data: Dataframe to which the solar features have to be added
        pj: prediction job which should at least contain the latitude and longitude location.
        feature_names: List of requested features

    Returns:
        DataFrame same as input dataframe with extra columns for the added solar features

    """
    # If pj is none add solar features with Utrecht as default location
    if pj is None:
        logger.info(
            "No prediction job, default location will be used for additional radiation features."
        )
        pj = {"lat": DEFAULT_LAT, "lon": DEFAULT_LON}

    # If features is none add solar feature anyway
    if feature_names is None:
        additional_solar_features = True

    # Otherwise check if they are among the requested features
    else:
        additional_solar_features = any(x in ["dni", "gti"] for x in feature_names)

    # Add add_additional_solar_features
    if "radiation" in data.columns and additional_solar_features:
        data["dni"] = calculate_dni(data["radiation"], pj)
        data["gti"] = calculate_gti(data["radiation"], pj)

    return data



# SPDX-FileCopyrightText: 2017-2023 Contributors to the OpenSTEF project <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0
"""Specifies the prediction job dataclass."""
from typing import Optional, Union

from pydantic import BaseModel

from openstef.data_classes.model_specifications import ModelSpecificationDataClass
from openstef.data_classes.split_function import SplitFuncDataClass
from openstef.data_classes.data_prep import DataPrepDataClass
from openstef.enums import PipelineType


class PredictionJobDataClass(BaseModel):
    """Holds all information about the specific forecast that has to be made."""

    id: Union[int, str]
    """The predictions job id (often abreviated as pid)."""
    model: str
    """The model type that should be used.

    Options are:
        - ``"xgb"``
        - ``"xgb_quantile"``
        - ``"lgb"``
        - ``"linear"``
        - ``"proloaf"`` (extra dependencies requiered, see README)

    If unsure what to pick, choose ``"xgb"``.

    """
    forecast_type: str
    """The type of forecasts that should be made.

    Options are:
        - ``"demand"``
        - ``"wind"``
        - ``"basecase"``

    If unsure what to pick, choose ``"demand"``.

    """
    horizon_minutes: int
    """The horizon of the desired forecast in minutes."""
    resolution_minutes: int
    """The resolution of the desired forecast in minutes."""
    lat: float
    """Latitude of the forecasted location in degrees."""
    lon: float
    """Longitude of the forecasted location in degrees."""
    name: str
    """Name of the forecast, e.g. the location name."""
    train_components: Optional[bool]
    """Whether splitting the forecasts in wind, solar, rest is desired."""
    description: Optional[str]
    """Optional description of the prediction job for human reference."""
    quantiles: Optional[list[float]]
    """Quantiles that have to be forecasted."""
    train_split_func: Optional[SplitFuncDataClass]
    """Optional custom splitting function for operational procces."""
    backtest_split_func: Optional[SplitFuncDataClass]
    """Optional custom splitting function for backtesting."""
    train_horizons_minutes: Optional[list[int]]
    """List of horizons that should be taken into account during training."""
    default_modelspecs: Optional[ModelSpecificationDataClass]
    """Default model specifications"""
    save_train_forecasts: bool = False
    """Indicate wether the forecasts produced during the training process should be saved."""
    completeness_treshold: float = 0.5
    """Minimum fraction of data that should be available for making a regular forecast."""
    minimal_table_length: int = 100
    """Minimum length (in rows) of the forecast input for making a regular forecast."""
    flatliner_treshold: int = 24
    """Number of consecutive values that have to be constant to detect a flatliner. """
    depends_on: Optional[list[Union[int, str]]]
    """Link to another prediction job on which this prediction job might depend."""
    sid: Optional[str]
    """Only required for create_solar_forecast task"""
    turbine_type: Optional[str]
    """Only required for create_wind_forecast task"""
    n_turbines: Optional[float]
    """Only required for create_wind_forecast task"""
    hub_height: Optional[float]
    """Only required for create_wind_forecast task"""
    pipelines_to_run: list[PipelineType] = [
        PipelineType.TRAIN,
        PipelineType.HYPER_PARMATERS,
        PipelineType.FORECAST,
    ]
    """The pipelines to run for this pj"""
    alternative_forecast_model_pid: Optional[Union[int, str]]
    """The pid that references another prediction job from which the model should be used for making forecasts."""
    data_prep_class: Optional[DataPrepDataClass]
    """The import string for the custom data prep class"""

    class Config:
        """Pydantic model configuration.

        This following configuration is needed to prevent ids in "depends_on" to be converted from int to str when we
        use integer ids.

        """

        smart_union = True

    def __getitem__(self, item: str) -> any:
        """Allows us to use subscription to get the items from the object."""
        return getattr(self, item)

    def __setitem__(self, key: str, value: any) -> None:
        """Allows us to use subscription to set the items in the object."""
        if hasattr(self, key):
            self.__dict__[key] = value
        else:
            raise AttributeError(f"{key} not an attribute of prediction job.")

    def get(self, key: str, default: any = None) -> any:
        """Allows to use the get functions similar to a python dict."""
        if hasattr(self, key):
            return getattr(self, key)
        else:
            return default



# SPDX-FileCopyrightText: 2017-2023 Contributors to the OpenSTEF project <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0
from datetime import datetime

import pandas as pd
import scipy.ndimage as mnts


def generate_forecast_datetime_range(
    forecast_data: pd.DataFrame,
) -> tuple[datetime, datetime]:
    """Generate forecast range based on last cluster of null values in first target column of forecast data.

    Example:

        A forecast dataset with data between 2021-11-05 and 2021-11-19, and the
        target column 'load' as first column is given as input to this function. The first
        column 'load' has null values between 2021-11-17 04:00:00 and 2021-11-19 05:00:00.
        The null values at the end of the column indicate when forecasts are needed.
        Therefore this function sets starting time of forecasts as 2021-11-17 04:00:00 and
        end time of forecasts as 2021-11-19 05:00:00.

    Args:
        forecast_data: The forecast dataframe.

    Returns:
        Start and end datetimes of the forecast range.

    """
    # By labeling the True/False values (based on the isnull() statement) as clusters,
    # we find what True value belongs to what cluster and the amount of True clusters.
    label_clusters, n_clusters = mnts.label(forecast_data.iloc[:, 0].isnull().values)

    # If there are zero true clusters, it means the target column does not have nulls
    if n_clusters == 0:
        raise ValueError(
            "Forecast target column must have null values to indicate "
            "when forecast starts and ends."
        )

    # If there are multiple true clusters, we select the last cluster as forecast range
    forecast_range_data = forecast_data.loc[label_clusters == n_clusters]
    # We select first datetime index of last cluster
    forecast_start_dt = forecast_range_data.index[0].to_pydatetime()

    # Forecast end is based on last datetime of given forecast data
    forecast_end_dt = forecast_data.index[-1].to_pydatetime()
    return forecast_start_dt, forecast_end_dt



# SPDX-FileCopyrightText: 2017-2023 Alliander N.V. <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0

"""This module contains historic load related functions used for feature engineering."""

import pandas as pd
import structlog

from openstef.data_classes.prediction_job import PredictionJobDataClass


def add_historic_load_as_a_feature(
    data: pd.DataFrame,
    pj: PredictionJobDataClass = None,
) -> pd.DataFrame:
    """Adds additional proloaf features to the input data, historic_load (equal to the load).

    Args:
        data: Dataframe to which the wind features have to be added
        pj: Prediction job.

    Returns:
        DataFrame that is the same as input dataframe with extra columns for the added proloaf features

    """
    logger = structlog.get_logger(__name__)

    if pj is None:
        pj = {}

    if pj.get("model") == "proloaf":
        data["historic_load"] = data["load"]
        logger.warning(
            "The historic_load is added to the data, this is a copy of the load. Adding"
            " this feature is in most of the cases not a good idea, it is designed for"
            " the proloaf model."
        )

    return data



# SPDX-FileCopyrightText: 2017-2021 Alliander N.V. <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0
"""This module defines the grouped regressor."""
from typing import Any, Callable, Union

import joblib
import numpy as np
import pandas as pd
from pandas.core.groupby.generic import DataFrameGroupBy
from sklearn.base import BaseEstimator, MetaEstimatorMixin, RegressorMixin, clone
from sklearn.utils.validation import check_is_fitted


class GroupedRegressor(BaseEstimator, RegressorMixin, MetaEstimatorMixin):
    """Meta-model that trains an instance of the base estimator for each key of a groupby operation applied on the data.

    The base estimator is a sklearn regressor, the groupby is performed on the columns specified in parameters.
    Moreover fit and predict methods can be performed in parallel for each group key thanks to joblib.

    Example:

    .. code-block:: md

        data =  | index | group | x0 | x1 | x3 | y |
                |   0   |   1   | .. | .. | .. | . |
                |   1   |   2   | .. | .. | .. | . |
                |   2   |   1   | .. | .. | .. | . |
                |   3   |   2   | .. | .. | .. | . |

                [              X              ][ Y ]


    The GroupedRegressor on the data with the group_columns='group' fits 2 models:
        - The model 1 with the row 0 and 2, columns x0, x1 and x3 as the features and column y as the target.
        - The model 2 with the row 1 and 3, columns x0, x1 and x3 as the features and column y as the target.

    Args:
        base_estimator: Regressor .

        group_columns: Name(s) of the column(s) used as the key for groupby operation.

        n_jobs: default=1
            The maximum number of concurrently running jobs,
            such as the number of Python worker processes when backend=”multiprocessing”
            or the size of the thread-pool when backend=”threading

    Attributes:
        feature_names_:  All input feature (without group_columns).

        estimators_:
            Dictionnary that stocks fitted estimators for each group.
            The keys are the keys of grouping and the values are the regressors fitted on the grouped data.

    """

    def __init__(
        self,
        base_estimator: RegressorMixin,
        group_columns: Union[str, int, list[str], list[int]],
        n_jobs: int = 1,
    ):
        """Initialize meta model."""
        self.base_estimator = base_estimator
        if type(group_columns) in [int, str]:
            self.group_columns = [group_columns]
        else:
            self.group_columns = group_columns
        self.n_jobs = n_jobs

    def _get_tags(self):
        return self.base_estimator._get_tags()

    def _check_group_columns(self, df):
        if type(self.group_columns) is not list:
            raise ValueError(
                "The group columns parameter should be a list, it gets a {}".format(
                    type(self.group_columns)
                )
            )
        for c in self.group_columns:
            if c not in list(df.columns):
                raise ValueError("The group column {} is missing!".format(c))

    def _partial_fit(
        self, group: Any, df_group: pd.DataFrame, eval_set=None, **kwargs
    ) -> tuple[Any, BaseEstimator]:
        estimator = clone(self.base_estimator)
        X = df_group.loc[:, self.feature_names_]
        y = df_group.loc[:, "__target__"]

        if eval_set is None:
            estimator_fitted = estimator.fit(X, y, **kwargs)
        else:
            estimator_fitted = estimator.fit(
                X,
                y,
                eval_set=[
                    (
                        df.loc[
                            (df[self.group_columns] == group).to_numpy().flatten(),
                            self.feature_names_,
                        ],
                        df.loc[
                            (df[self.group_columns] == group).to_numpy().flatten(),
                            "__target__",
                        ],
                    )
                    for df in eval_set
                ],
                **kwargs
            )

        return (group, estimator_fitted)

    def _partial_predict(self, group, df_group, **kwargs):
        return self.estimators_[group].predict(df_group, **kwargs)

    @classmethod
    def grouped_compute(
        cls,
        df: pd.DataFrame,
        group_columns: Union[list[str], list[int]],
        func: Callable[[tuple, pd.DataFrame], np.array],
        n_jobs: int = 1,
        eval_set=None,
    ) -> tuple[tuple[np.array, ...], DataFrameGroupBy, pd.DataFrame]:
        """Computes the specified function on each group defined by the grouping columns.

        It is an utility function used to perform fit and predict on each group.
        The df_res is the final dataframe that aggregate the results for each
        group. The group_res is a tuple where each field is corresponding to a results for a group. The gb is the
        grouping object.

        Args:
            df: DataFrame containing the input data necessary for the computation .
            group_columns: List of the columns used for the groupby operation
            func: Function that take the group key and the conrresponding data of this group
                and perform the computation on this group.
            n_jobs: The maximum number of concurrently running jobs,

        Returns:
            The tuple of the results of each group, the grouping dataframe and the global dataframe of results.

        """
        index_name = df.index.name or "index"
        df_reset = df.reset_index()

        df_res = pd.DataFrame(index=df_reset.index)

        gb = df_reset.groupby(group_columns)

        if n_jobs > 1:
            # Preferred scaling is at cluster level (e.g. k8s/serverless) instead of process level
            group_res = joblib.Parallel(n_jobs=n_jobs)(
                joblib.delayed(func)(
                    group, df_group.set_index(index_name).drop(group_columns, axis=1)
                )
                for group, df_group in gb
            )
        else:
            group_res = (
                func(group, df_group.set_index(index_name).drop(group_columns, axis=1))
                for group, df_group in gb
            )
        return group_res, gb, df_res

    def _grouped_predict(self, df: pd.DataFrame, n_jobs: int = 1, **kwargs) -> np.array:
        group_res, gb, df_res = self.grouped_compute(
            df,
            self.group_columns,
            lambda group, df_group: self._partial_predict(group, df_group, **kwargs),
            n_jobs,
        )

        for (group, group_index), result in zip(gb.groups.items(), group_res):
            df_res.loc[group_index, "__result__"] = np.array(result)

        return df_res["__result__"].to_numpy()

    def _grouped_fit(
        self, df: pd.DataFrame, n_jobs: int = 1, eval_set=None, **kwargs
    ) -> dict[Any, BaseEstimator]:
        group_res, _, _ = self.grouped_compute(
            df,
            self.group_columns,
            lambda group, df_group: self._partial_fit(
                group, df_group, eval_set=eval_set, **kwargs
            ),
            n_jobs,
        )
        return dict(group_res)

    def fit(self, x: np.ndarray, y: np.ndarray, eval_set=None, **kwargs):
        """Fit the model."""
        df = pd.DataFrame(x).copy(deep=True)
        self._check_group_columns(df)

        eval_df = None
        if eval_set is not None:
            eval_df = []
            for x_set, y_set in eval_set:
                self._check_group_columns(x_set)
                df_set = pd.DataFrame(x_set).copy(deep=True)
                df_set["__target__"] = y_set
                eval_df.append(df_set)

        self.feature_names_ = [
            c for c in list(df.columns) if c not in self.group_columns
        ]
        df.loc[:, "__target__"] = y
        self.estimators_ = self._grouped_fit(
            df, self.n_jobs, eval_set=eval_df, **kwargs
        )
        return self

    def predict(self, x: np.ndarray, **kwargs) -> np.ndarray:
        """Make a predicion."""
        check_is_fitted(self)
        df = pd.DataFrame(x)
        self._check_group_columns(df)
        return self._grouped_predict(df, self.n_jobs, **kwargs)



# SPDX-FileCopyrightText: 2017-2023 Contributors to the OpenSTEF project <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0

from lightgbm import LGBMRegressor

from openstef.model.regressors.regressor import OpenstfRegressor


class LGBMOpenstfRegressor(LGBMRegressor, OpenstfRegressor):
    """LGBM Regressor which implements the Openstf regressor API."""

    gain_importance_name = "gain"
    weight_importance_name = "split"

    @property
    def feature_names(self):
        return self._Booster.feature_name()

    @property
    def can_predict_quantiles(self):
        return False

    @staticmethod
    def _get_importance_names():
        return {
            "gain_importance_name": "gain",
            "weight_importance_name": "split",
        }



# SPDX-FileCopyrightText: 2017-2023 Contributors to the OpenSTEF project <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0
import os
from typing import Any

import optuna
import pandas as pd
import structlog

from openstef.data_classes.model_specifications import ModelSpecificationDataClass
from openstef.data_classes.prediction_job import PredictionJobDataClass
from openstef.exceptions import (
    InputDataInsufficientError,
    InputDataWrongColumnOrderError,
)
from openstef.feature_engineering.feature_applicator import TrainFeatureApplicator
from openstef.metrics.reporter import Report, Reporter
from openstef.model.model_creator import ModelCreator
from openstef.model.objective import RegressorObjective
from openstef.model.objective_creator import ObjectiveCreator
from openstef.model.regressors.regressor import OpenstfRegressor
from openstef.model.serializer import MLflowSerializer
from openstef.pipeline.train_model import (
    DEFAULT_TRAIN_HORIZONS,
    train_model_pipeline_core,
)
from openstef.validation import validation
from openstef.model_selection.model_selection import split_data_train_validation_test

optuna.logging.enable_propagation()  # Propagate logs to the root logger.
optuna.logging.disable_default_handler()  # Stop showing logs in sys.stderr.

logger = structlog.get_logger(__name__)

# See https://optuna.readthedocs.io/en/stable/reference/generated/optuna.study.Study.html#optuna.study.Study.optimize
N_TRIALS: int = 100  # The number of trials.
TIMEOUT: int = 600  # Stop study after the given number of second(s).


def optimize_hyperparameters_pipeline(
    pj: PredictionJobDataClass,
    input_data: pd.DataFrame,
    mlflow_tracking_uri: str,
    artifact_folder: str,
    horizons: list[float] = DEFAULT_TRAIN_HORIZONS,
    n_trials: int = N_TRIALS,
) -> dict:
    """Optimize hyperparameters pipeline.

    Expected prediction job key's: "name", "model"

    Args:
        pj: Prediction job
        input_data: Raw training input data
        mlflow_tracking_uri: Path/Uri to mlflow service
        artifact_folder: Path where artifacts, such as trained models, are stored
        horizons: horizons for feature engineering.
        n_trials: The number of trials. Defaults to N_TRIALS.

    Raises:
        ValueError: If the input_date is insufficient.

    Returns:
        Optimized hyperparameters.

    """
    (
        best_model,
        model_specs,
        report,
        trials,
        best_trial_number,
        best_params,
    ) = optimize_hyperparameters_pipeline_core(pj, input_data, horizons, n_trials)

    # Create serializer
    serializer = MLflowSerializer(mlflow_tracking_uri=mlflow_tracking_uri)

    # Save model, optimization results and report
    serializer.save_model(
        model=best_model,
        experiment_name=str(pj["id"]),
        model_type=pj["model"],
        model_specs=model_specs,
        report=report,
        phase="Hyperparameter_opt",
        trials=trials,
        trial_number=best_trial_number,
    )
    if artifact_folder:
        report_folder = os.path.join(artifact_folder, str(pj["id"]))
        Reporter.write_report_to_disk(report=report, report_folder=report_folder)
    return best_params


def optimize_hyperparameters_pipeline_core(
    pj: PredictionJobDataClass,
    input_data: pd.DataFrame,
    horizons: list[float] = DEFAULT_TRAIN_HORIZONS,
    n_trials: int = N_TRIALS,
) -> tuple[
    OpenstfRegressor, ModelSpecificationDataClass, Report, dict, int, dict[str, Any]
]:
    """Optimize hyperparameters pipeline core.

    Expected prediction job key's: "name", "model"

    Args:
        pj: Prediction job
        input_data: Raw training input data
        horizons: horizons for feature engineering.
        n_trials: The number of trials. Defaults to N_TRIALS.

    Raises:
        ValueError: If the input_date is insufficient.

    Returns:
        - Best model,
        - Model specifications of the best model,
        - Report of the best training round,
        - Trials,
        - Best trial number,
        - Optimized hyperparameters.

    """
    if pj.train_horizons_minutes is not None:
        horizons = pj.train_horizons_minutes

    if input_data.empty:
        raise InputDataInsufficientError("Input dataframe is empty")
    elif "load" not in input_data.columns:
        raise InputDataWrongColumnOrderError(
            "Missing the load column in the input dataframe"
        )

    # Validate and clean data
    validated_data = validation.drop_target_na(
        validation.validate(pj["id"], input_data, pj["flatliner_treshold"])
    )

    # Check if sufficient data is left after cleaning
    if not validation.is_data_sufficient(
        validated_data, pj["completeness_treshold"], pj["minimal_table_length"]
    ):
        raise InputDataInsufficientError(
            f"Input data is insufficient for {pj['name']} after validation and cleaning"
        )

    if pj.default_modelspecs:
        feature_names = pj.default_modelspecs.feature_names
        feature_modules = pj.default_modelspecs.feature_modules
    else:
        feature_names = None
        feature_modules = []

    if isinstance(horizons, str):
        if horizons not in set(input_data.columns):
            raise ValueError(
                f"The horizon parameter specifies a column name ({horizons}) missing in"
                " the input data."
            )
        else:
            # sort data to avoid same date repeated multiple time
            input_data = input_data.sort_values(horizons)

    validated_data_with_features = TrainFeatureApplicator(
        horizons=horizons, feature_names=feature_names, feature_modules=feature_modules
    ).add_features(validated_data, pj=pj)

    # Adds additional proloaf features to the input data, historic_load (equal to the load, first column)
    if pj["model"] == "proloaf" and "historic_load" not in list(
        validated_data_with_features.columns
    ):
        validated_data_with_features[
            "historic_load"
        ] = validated_data_with_features.iloc[:, 0]
        # Make sure horizons is last column
        temp_cols = validated_data_with_features.columns.tolist()
        new_cols = temp_cols[:-2] + [temp_cols[-1]] + [temp_cols[-2]]
        validated_data_with_features = validated_data_with_features[new_cols]

    # Create objective (NOTE: this is a callable class)
    objective = ObjectiveCreator.create_objective(model_type=pj["model"])

    study, objective = optuna_optimization(
        pj, objective, validated_data_with_features, n_trials
    )

    best_hyperparams = study.best_params
    best_model = study.user_attrs["best_model"]

    logger.info(
        f"Finished hyperparameter optimization, error objective {study.best_value} "
        f"and params {best_hyperparams}"
    )

    # Add quantiles to hyperparams so they are stored with the model info
    if pj["quantiles"]:
        best_hyperparams.update(quantiles=pj["quantiles"])

    # model specification
    model_specs = ModelSpecificationDataClass(
        id=pj["id"],
        feature_names=list(validated_data_with_features.columns),
        hyper_params=best_hyperparams,
    )

    # Train a model using the regular train pipeline.
    # The train/validation/test split used in hyperparam optimisation
    # is less suitable for an operational model.
    best_model, report, modelspecs, _ = train_model_pipeline_core(
        pj=pj, input_data=input_data, model_specs=model_specs
    )

    # Save model and report. Report is always saved to MLFlow and optionally to disk
    report = objective.create_report(model=best_model)

    trials = objective.get_trial_track()
    best_trial_number = study.best_trial.number

    return best_model, model_specs, report, trials, best_trial_number, study.best_params


def optuna_optimization(
    pj: PredictionJobDataClass,
    objective: RegressorObjective,
    validated_data_with_features: pd.DataFrame,
    n_trials: int,
) -> tuple[optuna.study.Study, RegressorObjective]:
    """Perform hyperparameter optimization with optuna.

    Args:
        pj: Prediction job
        objective: Objective function for optuna
        validated_data_with_features: cleaned input dataframe
        n_trials: number of optuna trials

    Returns:
        - Optimization study from optuna
        - The objective object used by optuna

    """
    model = ModelCreator.create_model(pj["model"])
    # Apply set to default hyperparameters if they are specified in the pj
    if pj.default_modelspecs:
        valid_hyper_parameters = {
            key: value
            for key, value in pj.default_modelspecs.hyper_params.items()
            if key in model.get_params().keys()
        }
        model.set_params(**valid_hyper_parameters)

    study = optuna.create_study(
        study_name=pj["model"],
        pruner=optuna.pruners.MedianPruner(n_warmup_steps=5),
        direction="minimize",
    )

    # Start with evaluating the default set of parameters,
    # this way the optimization never get worse than the default values
    study.enqueue_trial(objective.get_default_values())

    if pj.train_split_func is None:
        split_func = split_data_train_validation_test
        split_args = {
            "stratification_min_max": pj["model"] != "proloaf",
            "back_test": True,
        }
    else:
        split_func, split_args = pj.train_split_func.load(
            required_arguments=["data", "test_fraction", "validation_fraction"]
        )

    objective = objective(
        model,
        validated_data_with_features,
        split_func=split_func,
        split_args=split_args,
    )

    # Optuna updates the model by itself
    # and the model is the optimized over this finishes
    study.optimize(
        objective,
        n_trials=n_trials,
        callbacks=[_log_study_progress_and_save_best_model],
        show_progress_bar=False,
        timeout=TIMEOUT,
    )

    return study, objective


def _log_study_progress_and_save_best_model(
    study: optuna.study.Study, trial: optuna.trial.FrozenTrial
) -> None:
    # Collect study and trial data
    # trial_index = study.trials.index(trial)
    # best_trial_index = study.trials.index(study.best_trial)
    value = trial.value
    params = trial.params
    duration = (trial.datetime_complete - trial.datetime_start).total_seconds()
    # Log information about this trial
    logger.debug(
        f"Trial {trial.number} finished with value: {value} and parameters: {params}."
        f"Best trial is {study.best_trial.number}. Iteration took {duration} s"
    )
    # If this trial is the best save the model as a study attribute
    if study.best_trial.number == trial.number:
        study.set_user_attr(key="best_model", value=trial.user_attrs["model"])



# SPDX-FileCopyrightText: 2017-2023 Contributors to the OpenSTEF project <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0
from enum import Enum


# TODO replace this with ModelType (MLModelType == Machine Learning model type)
class MLModelType(Enum):
    XGB = "xgb"
    XGB_QUANTILE = "xgb_quantile"
    LGB = "lgb"
    LINEAR = "linear"
    ProLoaf = "proloaf"
    ARIMA = "arima"


class ForecastType(Enum):
    DEMAND = "demand"
    WIND = "wind"
    SOLAR = "solar"
    BASECASE = "basecase"


class TracyJobResult(Enum):
    SUCCESS = "success"
    FAILED = "failed"
    UNKNOWN = "unknown"


class PipelineType(Enum):
    FORECAST = "forecast"
    TRAIN = "train"
    HYPER_PARMATERS = "hyper_parameters"



# SPDX-FileCopyrightText: 2017-2023 Contributors to the OpenSTEF project <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0

"""Openstef custom exceptions."""
# Define custom exception
class NoPredictedLoadError(Exception):
    """No predicted load for given datatime range."""

    def __init__(
        self,
        pid: int,
        message: str = "No predicted load found",
    ):
        self.pid = pid
        self.message = message
        super().__init__(self.message)


class NoRealisedLoadError(Exception):
    """No realised load for given datetime range."""

    def __init__(
        self,
        pid: int,
        message: str = "No realised load found",
    ):
        self.pid = pid
        self.message = message
        super().__init__(self.message)


class InputDataInvalidError(Exception):
    """Invalid input data."""


class InputDataInsufficientError(InputDataInvalidError):
    """Insufficient input data."""


class InputDataWrongColumnOrderError(InputDataInvalidError):
    """Wrong column order input data."""


class OldModelHigherScoreError(Exception):
    """Old model has a higher score then new model."""


class ModelWithoutStDev(Exception):
    """A machine learning model should have a valid standard deviation."""


class ComponentForecastTooShortHorizonError(Exception):
    """Component forecasts should be available for at least 30 hours in advance."""


class PredictionJobException(Exception):
    """One or more prediction jobs raised an exception."""

    def __init__(self, metrics=None):
        super().__init__("One or more prediction jobs raised an exception.")
        if metrics is None:
            metrics = {}
        self.metrics = metrics


class SkipSaveTrainingForecasts(Exception):
    """If old model is better or too young, you don't need to save the traing forcast."""



# SPDX-FileCopyrightText: 2017-2023 Contributors to the OpenSTEF project <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0
"""This module contains the CRON job that is periodically executed to make prognoses of solar features.

These are useful for splitting the load in solar and wind contributions.

Example:
    This module is meant to be called directly from a CRON job. A description of
    the CRON job can be found in the /k8s/CronJobs folder.
    Alternatively this code can be run directly by running::
        $ python create_solar_forecast

"""
from datetime import datetime, timedelta
from pathlib import Path

import numpy as np
import pandas as pd
from scipy import optimize

from openstef import PROJECT_ROOT
from openstef.tasks.utils.predictionjobloop import PredictionJobLoop
from openstef.tasks.utils.taskcontext import TaskContext

# TODO move to config
PV_COEFS_FILEPATH = PROJECT_ROOT / "openstef" / "data" / "pv_single_coefs.csv"


def make_solar_prediction_pj(pj, context, radius=30, peak_power=180961000.0):
    """Make a solar prediction for a specific prediction job.

    Args:
        pj: (dict) prediction job
        context: Task context
        radius: Radius us to collect PV systems.
        peak_power: Peak power.

    """
    context.logger.info("Get solar input data from database")
    # pvdata is only stored in the prd database
    solar_input = context.database.get_solar_input(
        (pj["lat"], pj["lon"]),
        pj["horizon_minutes"],
        pj["resolution_minutes"],
        radius=radius,
        sid=pj["sid"],
    )

    if len(solar_input) == 0:
        raise ValueError("Empty solar input")

    context.logger.info("Make solar prediction using Fides")
    power = fides(
        solar_input[["aggregated", "radiation"]].rename(
            columns=dict(radiation="insolation", aggregated="load")
        )
    )

    # if the forecast is for a region, output should be scaled to peak power
    if (radius != 0) and (not np.isnan(peak_power)):
        power = peak_power / max(solar_input.aggregated) * power
    context.logger.info("Store solar prediction in database")
    power["pid"] = pj["id"]
    power["type"] = "solar"
    power["algtype"] = "Fides"
    power["customer"] = pj["name"]
    power["description"] = pj["description"]
    context.database.write_forecast(power)


def combine_forecasts(forecasts, combination_coefs):
    """This function combines several independent forecasts into one, using predetermined coefficients.

    Input:
        - forecasts: pd.DataFrame(index = datetime, algorithm1, ..., algorithmn)
        - combinationcoefs: pd.DataFrame(param1, ..., paramn, algorithm1, ..., algorithmn)

    Output:
        - pd.DataFrame(datetime, forecast)

    """
    models = [x for x in list(forecasts) if x not in ["created", "datetime"]]

    # Add subset parameters to df
    # Identify which parameters should be used to define subsets based on the
    # combination coefs
    subset_columns = [
        "tAhead",
        "hForecasted",
        "weekday",
        "hForecastedPer6h",
        "tAheadPer2h",
        "hCreated",
    ]
    subset_defs = [x for x in list(combination_coefs) if x in subset_columns]

    df = forecasts.copy()
    # Now add these subsetparams to df
    if "tAhead" in subset_defs:
        t_ahead = (df["datetime"] - df["created"]).dt.total_seconds() / 3600
        df["tAhead"] = t_ahead

    if "hForecasted" in subset_defs:
        df["hForecasted"] = df.datetime.dt.hour

    if "weekday" in subset_defs:
        df["weekday"] = df.datetime.dt.weekday

    if "hForecastedPer6h" in subset_defs:
        df["hForecastedPer6h"] = pd.to_numeric(
            np.floor(df.datetime.dt.hour / 6) * 6, downcast="integer"
        )

    if "tAheadPer2h" in subset_defs:
        df["tAheadPer2h"] = pd.to_numeric(
            np.floor((df.datetime - df.created).dt.total_seconds() / 60 / 60 / 2) * 2,
            downcast="integer",
        )

    if "hCreated" in subset_defs:
        df["hCreated"] = df.created.dt.hour

    # Start building combinationcoef dataframe that later will be multiplied with the
    # individual forecasts
    # This is the best way for a backtest:
    #    uniquevalues = list([np.unique(df[param].values) for param in subsetDefs])
    #    permutations = list(itertools.product(*uniquevalues))

    # This is the best way for a single forecast
    permutations = [tuple(x) for x in df[subset_defs].values]

    result_df = pd.DataFrame()

    for subsetvalues in permutations:
        subset = df.copy()
        coefs = combination_coefs

        # Create subset based on all subsetparams, for forecasts and coefs
        for value, param in zip(subsetvalues, subset_defs):
            subset = subset.loc[subset[param] == value]
            # Define function which find closest match of a value from an array of values.
            #  Use this later to find best coefficient from the given subsetting dividers
            closest_match = min(coefs[param], key=lambda x, val=value: abs(x - val))
            coefs = coefs.loc[coefs[param] == closest_match]
            # Find closest matching value for combinationCoefParams corresponding to
            # available subsetValues

        # Of course, not all possible subsets have to be defined in the forecast.
        # Skip empty subsets
        if len(subset) == 0:
            continue

        # Multiply forecasts with their coefficients
        result = np.multiply(subset[models], np.array(coefs[models]))
        result["forecast"] = result.apply(np.nansum, axis=1)
        # Add handling with NA values for a single forecast
        result["coefsum"] = np.nansum(coefs[models].values)
        nanselector = np.isnan(subset[models].iloc[0].values)
        result["nonnacoefsum"] = np.nansum(coefs[models].values.flatten() * nanselector)
        result["forecast"] = (
            result["forecast"]
            * result["coefsum"]
            / (result["coefsum"] - result["nonnacoefsum"])
        )
        result["datetime"] = subset["datetime"]
        result["created"] = subset["created"]
        result = result[["datetime", "created", "forecast"]]
        result_df = pd.concat([result_df, result])
    # sort by datetime
    result_df.sort_values(["datetime", "created"], inplace=True)

    return result_df


def fides(data: pd.DataFrame, all_forecasts: bool = False):
    """Fides makes a forecast based on persistence and a direct fit with insolation.

    Args:
        data: pd.DataFrame(index = datetime, columns =['output','insolation'])
        all_forecasts: Should all forecasts be returned or only the combination

    Example:
    import numpy as np
    index = pd.date_range(start = "2017-01-01 09:00:00", freq = '15T', periods = 300)
    data = pd.DataFrame(index = index,
                        data = dict(load=np.sin(index.hour/24*np.pi)*np.random.uniform(0.7,1.7, 300)))
    data['insolation'] = data.load * np.random.uniform(0.8, 1.2, len(index)) + 0.1
    data.loc[int(len(index)/3*2):,"load"] = np.NaN

    """
    insolation_forecast = apply_fit_insol(data, add_to_df=False)
    persistence = apply_persistence(data, how="mean", smooth_entries=4, add_to_df=True)

    df = insolation_forecast.merge(persistence, left_index=True, right_index=True)

    coefs = pd.read_csv(PV_COEFS_FILEPATH)

    # Apply combination coefs
    df["created"] = df.loc[df.load.isnull()].index.min()
    forecast = combine_forecasts(
        df.loc[df.load.isnull(), ["forecaopenstefitInsol", "persistence", "created"]]
        .reset_index()
        .rename(columns=dict(index="datetime")),
        coefs,
    ).set_index("datetime")[["forecast"]]

    if all_forecasts:
        forecast = forecast.merge(
            df[["persistence", "forecaopenstefitInsol"]],
            left_index=True,
            right_index=True,
            how="left",
        )

    return forecast


def main(config=None, database=None):
    taskname = Path(__file__).name.replace(".py", "")

    if database is None or config is None:
        raise RuntimeError(
            "Please specify a config object and/or database connection object. These"
            " can be found in the openstef-dbc package."
        )

    with TaskContext(taskname, config, database) as context:
        context.logger.info("Querying solar prediction jobs from database")
        prediction_jobs = context.database.get_prediction_jobs_solar()
        num_prediction_jobs = len(prediction_jobs)

        # only make customer = Provincie once an hour
        utc_now_minute = datetime.utcnow().minute
        if utc_now_minute >= 15:
            prediction_jobs = [
                pj for pj in prediction_jobs if str(pj["name"]).startswith("Provincie")
            ]
            num_removed_jobs = num_prediction_jobs - len(prediction_jobs)
            num_prediction_jobs = len(prediction_jobs)
            context.logger.info(
                "Remove 'Provincie' solar predictions",
                num_removed_jobs=num_removed_jobs,
                num_prediction_jobs=num_prediction_jobs,
            )

        PredictionJobLoop(context, prediction_jobs=prediction_jobs).map(
            make_solar_prediction_pj, context
        )


def calc_norm(data, how="max", add_to_df=True):
    """This script calculates the norm of a given dataset.

    Input:
        - data: pd.DataFrame(index = datetime, columns = [load])
        - how: str can be any function from numpy, recognized by np.'how'
        Optional:
        - add_to_df: Bool, add the norm to the data

    Output:
        - pd.DataFrame(index = datetime, columns = [load])
    NB: range of datetime of input is equal to range of datetime of output

    Example:
    import pandas as pd
    import numpy as np
    index = pd.date_range(start = "2017-01-01 09:00:00", freq = '15T', periods = 200)
    data = pd.DataFrame(index = index,
                        data = dict(load=np.sin(index.hour/24*np.pi)*np.random.uniform(0.7,1.7, 200)))

    """
    colname = list(data)[0]
    if how == "max":
        df = data.groupby(data.index.time).apply(lambda x: x.max(skipna=True))
    if how == "mean":
        df = data.groupby(data.index.time).apply(lambda x: x.mean(skipna=True))

    # rename
    df.rename(columns={colname: "Norm"}, inplace=True)

    # Merge to dataframe if add_to_df == True
    if add_to_df:
        df = data.merge(df, left_on=data.index.time, right_index=True)[
            [colname, "Norm"]
        ].sort_index()

    return df


def apply_persistence(data, how="mean", smooth_entries=4, add_to_df=True, colname=None):
    """This script calculates the persistence forecast.

    Input:
        - data: pd.DataFrame(index = datetime, columns = [load]), datetime is expected to have historic values, as well as NA values
        Optional:
        - how: str, how to determine the norm (abs or mean)
        - smoothEntries: int, number of historic entries over which the persistence is smoothed
        - add_to_df: Bool, add the forecast to the data
        - option of specifying colname if load is not first column

    Output:
        - pd.DataFrame(index = datetime, columns = [(load,) persistence])
    NB: range of datetime of input is equal to range of datetime of output

    Example:
    import pandas as pd
    import numpy as np
    index = pd.date_range(start = "2017-01-01 09:00:00", freq = '15T', periods = 300)
    data = pd.DataFrame(index = index,
                        data = dict(load=np.sin(index.hour/24*np.pi)*np.random.uniform(0.7,1.7, 300)))
    data.loc[200:,"load"] = np.nan

    """
    data = data.sort_index()

    if colname is None:
        colname = list(data)[0]

    df = calc_norm(data, how=how, add_to_df=True)

    # this selects the last non NA values
    last_entries = df.loc[df[colname].notnull()][-smooth_entries:]

    norm_mean = last_entries.Norm.mean()
    if norm_mean == 0:
        norm_mean = 1

    factor = last_entries[colname].mean() / norm_mean
    df["persistence"] = df.Norm * factor

    if add_to_df:
        df = df[[colname, "persistence"]]
    else:
        df = df[["persistence"]]

    return df


def apply_fit_insol(data, add_to_df=True, hours_delta=None, polynomial=False):
    """This model fits insolation to PV yield and uses this fit to forecast PV yield. It uses a 2nd order polynomial.

    Input:
        - data: pd.DataFrame(index = datetime, columns = [load, insolation])
        Optional:
        - hoursDelta: period of forecast in hours [int] (e.g. every 6 hours for KNMI)
        - addToDF: Bool, add the norm to the data

    Output:
        - pd.DataFrame(index = datetime, columns = [(load), forecaopenstefitInsol])
    NB: range of datetime of input is equal to range of datetime of output

    Example:
    import pandas as pd
    import numpy as np
    index = pd.date_range(start = "2017-01-01 09:00:00", freq = '15T', periods = 300)
    data = pd.DataFrame(index = index,
                        data = dict(load=np.sin(index.hour/24*np.pi)*np.random.uniform(0.7,1.7, len(index))))
    data['insolation'] = data.load * np.random.uniform(0.8, 1.2, len(index)) + 0.1
    data.loc[int(len(index)/3*2):,"load"] = np.NaN

    """
    colname = list(data)[0]

    # Define subset, only keep non-NaN values and the most recent forecasts
    # This ensures a good training set
    if hours_delta is None:
        subset = data.loc[(data[colname].notnull()) & (data[colname] > 0)]
    else:
        subset = data.loc[
            (data[colname].notnull())
            & (data[colname] > 0)
            & (data["tAhead"] < timedelta(hours=hours_delta))
            & (data["tAhead"] >= timedelta(hours=0))
        ]

    def linear_fun(coefs, values):
        return coefs[0] * values + coefs[1]

    def second_order_poly(coefs, values):
        return coefs[0] * values**2 + coefs[1] * values + coefs[2]

    # Define function to be minimized and subsequently minimize this function
    if polynomial:
        # Define starting guess
        x0 = [1, 1, 0]  # ax**2 + bx + c.
        fun = (
            lambda x: (second_order_poly(x, subset.insolation) - subset[colname])
            .abs()
            .mean()
        )
        # , bounds = bnds, constraints = cons)
        res = optimize.minimize(fun, x0)
        # Apply fit
        df = second_order_poly(res.x, data[["insolation"]]).rename(
            columns=dict(insolation="forecaopenstefitInsol")
        )

    else:
        x0 = [1, 0]
        fun = (
            lambda x: (linear_fun(x, subset.insolation) - subset[colname]).abs().mean()
        )
        res = optimize.minimize(fun, x0)
        df = linear_fun(res.x, data[["insolation"]]).rename(
            columns=dict(insolation="forecaopenstefitInsol")
        )

    # Merge to dataframe if addToDF == True
    if add_to_df:
        if hours_delta is None:
            df = data.merge(df, left_index=True, right_index=True)
        else:
            df = pd.concat([data, df], axis=1)

    return df


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: 2017-2023 Contributors to the OpenSTEF project <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0
"""Specifies the split function dataclass."""
import inspect
import json
from importlib import import_module
from typing import Any, Sequence, Union, TypeVar

from pydantic import BaseModel

DataPrepClass = TypeVar("DataPrepClass")


class DataPrepDataClass(BaseModel):
    """Class that allows to specify a custom class to prepare the data (feature engineering , etc ...)."""

    klass: Union[str, type[DataPrepClass]]
    arguments: Union[
        str, dict[str, Any]
    ]  # JSON string holding the function parameters or dict

    def __getitem__(self, key: str):
        """Allows us to use subscription to get the items from the object."""
        return getattr(self, key)

    def __setitem__(self, key: str, value: any):
        """Allows us to use subscription to set the items in the object."""
        if hasattr(self, key):
            self.__dict__[key] = value
        else:
            raise AttributeError(f"{key} not an attribute of prediction job.")

    def _load_klass(
        self, required_arguments: Sequence[str] = None
    ) -> type[DataPrepClass]:
        """Load data prep class from path.

        Args:
            klass_path (str): The path to the data prep class

        Returns:
            klass (type[AbstractDataPreparation]): The loaded data prep class

        """
        if isinstance(self.klass, str):
            path_elements = self.klass.split(".")
            module_path = ".".join(path_elements[:-1])
            module = import_module(module_path)
            klass_name = path_elements[-1]
            klass = getattr(module, klass_name)
        else:
            klass = self.klass

        # Check that the klass accepts mandatory arguments
        if not inspect.isclass(klass):
            raise ValueError("The loaded object is not a class: {klass!r}")

        if required_arguments is not None:
            klass_params = set(inspect.signature(klass).parameters)

            if len(set(required_arguments) - klass_params) > 0:
                raise ValueError(
                    "The loaded data prep class does not have the required arguments"
                )

        return klass

    def _load_arguments(self) -> dict[str, Any]:
        """Load the arguments.

        Convert the arguments from JSON if they are given as strings or simply return them otherwise.

        Returns:
            arguments (dict[str, Any]): The additional arguments to be passed to the class

        """
        if isinstance(self.arguments, str):
            return json.loads(self.arguments)
        else:
            return self.arguments

    def load(
        self, required_arguments: Sequence[str] = None
    ) -> tuple[type[DataPrepClass], dict[str, Any]]:
        """Load the function and its arguments.

        If the function and the arguments are given as strings in the instane attributes, load them as Python objects
        otherwise just return them from the instance attributes.

        Args:
            required_arguments (list[str]): list of arguments the loaded class must have

        Returns:
            - class (type[AbstractDataPreparation])
            - arguments (dict[str, Any])

        """
        return self._load_klass(required_arguments), self._load_arguments()



# SPDX-FileCopyrightText: 2017-2023 Contributors to the OpenSTEF project <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0
import random
from typing import Iterable, Sequence, Set, Union

import networkx as nx

from openstef.data_classes.prediction_job import PredictionJobDataClass

NodeIdType = Union[str, int]
EdgeType = tuple[NodeIdType, NodeIdType]


def has_dependencies(pjs: Iterable[PredictionJobDataClass]) -> bool:
    """Test whether some prediction jobs have dependencies information.

    Args:
        pjs: The list of prediction jobs

    Returns:
        True if some dependency information was found.

    """
    for pj in pjs:
        if pj.depends_on is not None and len(pj.depends_on) > 0:
            return True
    return False


def build_graph_structure(
    pjs: Iterable[PredictionJobDataClass],
) -> tuple[Set[NodeIdType], Set[EdgeType]]:
    """Build the graph of dependencies between prediction jobs.

    Args:
        pjs: The Iterable of prediction jobs

    Returns:
        - The set of node ids of the graph
        - The set of edges in the graph

    """
    nodes = set()
    edges = set()

    for pj in pjs:
        nodes.add(pj["id"])
        if pj.depends_on is not None:
            for j in pj.depends_on:
                edges.add((j, pj["id"]))

    return nodes, edges


def build_nx_graph(
    nodes: Iterable[NodeIdType], edges: Iterable[EdgeType]
) -> nx.DiGraph:
    """Build a networkx Directed Graph.

    Args:
        nodes: The sequence of node ids
        edges: The sequence of edges

    Returns:
        The dependency graph

    """
    graph = nx.DiGraph()
    graph.add_nodes_from(nodes)
    graph.add_edges_from(edges)
    return graph


def find_groups(
    pjs: Sequence[PredictionJobDataClass], randomize_groups: bool = False
) -> tuple[nx.DiGraph, list[list[PredictionJobDataClass]]]:
    """Find a sequence of prediction job groups respecting dependencies.

    Compute groups of prediction jobs such that the prediction jobs in a group
    depend of at least one prediction job in the previous group and does not depend
    on a prediction job in the following groups.
    This means that all the prediction jobs in a group can be run in parallel and that
    if groups are treated in the given order, the dependencies of a prediction job have
    already been treated when the prediction job is run.

    Args:
        pjs: The sequence of prediction jobs
        randomize_groups: Wether subgroups should be randomized.

    Returns:
        - The dependency graph
        - The list of prediction job groups

    """
    nodes, edges = build_graph_structure(pjs)
    graph = build_nx_graph(nodes, edges)
    groups = list(nx.topological_generations(graph))

    if randomize_groups:
        for group in groups:
            random.shuffle(group)

    # Convert groups of pj ids to groups of pjs
    pj_id_map = {pj["id"]: i for i, pj in enumerate(pjs)}
    pj_groups = [[pjs[pj_id_map[pj_id]] for pj_id in group] for group in groups]
    return graph, pj_groups



# SPDX-FileCopyrightText: 2017-2023 Contributors to the OpenSTEF project <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0
from pathlib import Path

import pandas as pd
import structlog

from openstef.data_classes.prediction_job import PredictionJobDataClass
from openstef.exceptions import NoRealisedLoadError
from openstef.feature_engineering.feature_applicator import (
    OperationalPredictFeatureApplicator,
)
from openstef.model.basecase import BaseCaseModel
from openstef.model.confidence_interval_applicator import ConfidenceIntervalApplicator
from openstef.pipeline.utils import generate_forecast_datetime_range
from openstef.postprocessing.postprocessing import (
    add_components_base_case_forecast,
    add_prediction_job_properties_to_forecast,
)
from openstef.validation import validation

MODEL_LOCATION = Path(".")
BASECASE_HORIZON_MINUTES = 60 * 24 * 14  # 14 days ahead
BASECASE_RESOLUTION_MINUTES = 15


def create_basecase_forecast_pipeline(
    pj: PredictionJobDataClass,
    input_data: pd.DataFrame,
) -> pd.DataFrame:
    """Compute the base case forecast and confidence intervals for a given prediction job and input data.

    Args:
        pj: Prediction job
        input_data: data frame containing the input data necessary for the prediction.

    Returns:
        Basecase forecast

    """
    logger = structlog.get_logger(__name__)

    logger.info("Preprocessing data for basecase forecast")
    # Validate data
    # Currently effectively disabled by giving None.
    # We keep this step so it later can be filled using arguments defined in PJ
    validated_data = validation.validate(pj["id"], input_data, flatliner_threshold=None)

    # Add features
    data_with_features = OperationalPredictFeatureApplicator(
        horizons=[0.25],
        feature_names=[
            "T-7d",
            "T-14d",
        ],  # Generate features for load 7 days ago and load 14 days ago these are the same as the basecase forecast.
    ).add_features(validated_data)

    # Similarly to the forecast pipeline, only try to make a forecast for moments in the future
    # TODO, do we want to be this strict on time window of forecast in this place?
    # see issue https://github.com/OpenSTEF/openstef/issues/121
    forecast_start, forecast_end = generate_forecast_datetime_range(data_with_features)
    forecast_input = data_with_features[forecast_start:forecast_end]

    # Initialize model
    model = BaseCaseModel()
    logger.info("Making basecase forecast")
    # Make basecase forecast
    basecase_forecast = BaseCaseModel().predict(forecast_input)

    # Check if input data is available
    if len(basecase_forecast) == 0:
        raise NoRealisedLoadError(pj["id"])

    # Estimate the stdev by using the stdev of the hour for historic (T-14d) load
    model.standard_deviation = generate_basecase_confidence_interval(forecast_input)
    logger.info("Postprocessing basecase forecast")
    # Apply confidence interval
    basecase_forecast = ConfidenceIntervalApplicator(
        model, forecast_input
    ).add_confidence_interval(basecase_forecast, pj)

    # Add basecase for the component forecasts
    basecase_forecast = add_components_base_case_forecast(basecase_forecast)

    # Do further postprocessing
    basecase_forecast = add_prediction_job_properties_to_forecast(
        pj=pj,
        forecast=basecase_forecast,
        algorithm_type="basecase_lastweek",
        forecast_quality="not_renewed",
    )

    return basecase_forecast


def generate_basecase_confidence_interval(
    data_with_features: pd.DataFrame,
) -> pd.DataFrame:
    """Calculate confidence interval for a basecase forecast.

    Args:
        data_with_features: Input dataframe that is used to make the basecase forecast.

    Returns:
        Dataframe with the confidence interval.

    """
    confidence_interval = (
        data_with_features[["T-14d"]]  # Select only the T-14d column as a DataFrame
        .groupby(data_with_features.index.hour)  # Get the std for every hour
        .std()
        .rename(columns={"T-14d": "stdev"})  # Rename the column to stdev
    )
    confidence_interval["hour"] = confidence_interval.index
    confidence_interval["horizon"] = 48
    return confidence_interval



# SPDX-FileCopyrightText: 2017-2023 Contributors to the OpenSTEF project <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0
from abc import abstractmethod
from typing import Union

import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator, RegressorMixin


class OpenstfRegressor(BaseEstimator):
    """This class defines the interface to which all ML models within OpenSTEF should adhere.

    Required methods are indicated by abstractmethods, for which concrete implementations of ML models should have a
    definition. Common functionality which is required for the automated pipelines in OpenSTEF is defined in this class.

    """

    def __init__(self):
        self.feature_importance_dataframe = None
        self.feature_importances_ = None

    def score(self, X, y):
        """Makes `score` method from RegressorMixin available."""
        return RegressorMixin.score(self, X, y)

    ## Define abstract methods required to be implemented by concrete models
    @property
    @abstractmethod
    def feature_names(self) -> list:
        """Retrieve the model input feature names.

        Returns:
            The list of feature names

        """

    @property
    @abstractmethod
    def can_predict_quantiles(self) -> bool:
        """Attribute that indicates if the model predict particular quantiles.

        e.g. XGBQuantileOpenstfRegressor

        """

    @abstractmethod
    def predict(self, x: pd.DataFrame, **kwargs) -> np.array:
        """Makes a prediction. Only available after the model has been trained.

        Args:
            x: Feature matrix
            kwargs: model-specific keywords

        Returns:
            Prediction

        """

    @abstractmethod
    def fit(self, x: np.array, y: np.array, **kwargs) -> RegressorMixin:
        """Fits the regressor.

        Args:
            x: Feature matrix
            y: Labels
            kwargs: model-specific keywords

        Returns:
            Fitted model

        """

    def set_feature_importance(self) -> Union[pd.DataFrame, None]:
        """Get feature importance.

        Returns:
            DataFrame with feature importance.

        """
        # returns a dict if we can get feature importance else returns None
        importance_names = self._get_importance_names()
        # if the model doesn't support feature importance return None
        if importance_names is None:
            return None

        gain = self._fraction_importance(importance_names["gain_importance_name"])
        weight_importance = self._fraction_importance(
            importance_names["weight_importance_name"]
        )

        feature_importance = pd.DataFrame(
            {"gain": gain, "weight": weight_importance}, index=self.feature_names
        )

        feature_importance.sort_values(by="gain", ascending=False, inplace=True)
        return feature_importance

    def _fraction_importance(self, importance: str) -> np.ndarray:
        self.importance_type = importance
        feature_importance = self.feature_importances_
        feature_importance = feature_importance / sum(feature_importance)
        return feature_importance

    @staticmethod
    def _get_importance_names() -> Union[dict, None]:
        """Get importance names if applicable.

        Returns:
            A dict or None, return None if the model can't get feature importance

        """
        return None



# SPDX-FileCopyrightText: 2017-2023 Contributors to the OpenSTEF project <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0
"""This module should be executed once every day.

For all prediction_jobs, it will create a 'basecase' forecast which is less accurate, but (almost) always available.
For now, it uses the load a week earlier.
Missing datapoints are interpolated.

Example:
    This module is meant to be called directly from a CRON job. A description of the
    CRON job can be found in the /k8s/CronJobs folder.

    Alternatively this code can be run directly by running:

        $ python create_basecase_forecast.py

"""
from datetime import datetime, timedelta
from pathlib import Path

import pandas as pd

from openstef.data_classes.prediction_job import PredictionJobDataClass
from openstef.enums import PipelineType
from openstef.pipeline.create_basecase_forecast import create_basecase_forecast_pipeline
from openstef.tasks.utils.predictionjobloop import PredictionJobLoop
from openstef.tasks.utils.taskcontext import TaskContext

T_BEHIND_DAYS: int = 15
T_AHEAD_DAYS: int = 14


def create_basecase_forecast_task(
    pj: PredictionJobDataClass, context: TaskContext
) -> None:
    """Top level task that creates a basecase forecast.

    On this task level all database and context manager dependencies are resolved.

    Args:
        pj: Prediction job
        context: Contect object that holds a config manager and a database connection

    """
    # Check pipeline types
    if PipelineType.FORECAST not in pj.pipelines_to_run:
        context.logger.info(
            "Skip this PredictionJob because forecast pipeline is not specified in the pj."
        )
        return

    # Define datetime range for input data
    datetime_start = datetime.utcnow() - timedelta(days=T_BEHIND_DAYS)
    datetime_end = datetime.utcnow() + timedelta(days=T_AHEAD_DAYS)

    # Retrieve input data
    input_data = context.database.get_model_input(
        pid=pj["id"],
        location=[pj["lat"], pj["lon"]],
        datetime_start=datetime_start,
        datetime_end=datetime_end,
    )

    # Make basecase forecast using the corresponding pipeline
    basecase_forecast = create_basecase_forecast_pipeline(pj, input_data)

    # Do not store basecase forecasts for moments within next 48 hours.
    # Those should be updated by regular forecast process.
    basecase_forecast = basecase_forecast.loc[
        basecase_forecast.index
        > (pd.to_datetime(datetime.utcnow(), utc=True) + timedelta(hours=48)),
        :,
    ]

    # Write basecase forecast to the database
    context.database.write_forecast(basecase_forecast, t_ahead_series=True)


def main(config: object = None, database: object = None):
    taskname = Path(__file__).name.replace(".py", "")

    if database is None or config is None:
        raise RuntimeError(
            "Please specifiy a config object and/or database connection object. These"
            " can be found in the openstef-dbc package."
        )

    with TaskContext(taskname, config, database) as context:
        model_type = ["xgb", "xgb_quantile", "lgb"]

        PredictionJobLoop(context, model_type=model_type).map(
            create_basecase_forecast_task, context
        )


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: 2017-2023 Contributors to the OpenSTEF project <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0
"""This module defines several FeatureApplicators.

These applicatiors are used to add features to the input data in the corresponding pipelines.

"""
from abc import ABC, abstractmethod
from typing import Optional, Union

import numpy as np
import pandas as pd

from openstef.data_classes.prediction_job import PredictionJobDataClass
from openstef.feature_engineering.apply_features import apply_features
from openstef.feature_engineering.feature_adder import (
    FeatureDispatcher,
    adders_from_modules,
)
from openstef.feature_engineering.general import (
    add_missing_feature_columns,
    enforce_feature_order,
    remove_non_requested_feature_columns,
)

LATENCY_CONFIG = {"APX": 24}  # A specific latency is part of a specific feature.


class AbstractFeatureApplicator(ABC):
    """Defines the Applicator interface."""

    def __init__(
        self,
        horizons: Union[list[float], str],
        feature_names: Optional[list[str]] = None,
        feature_modules: Optional[list[str]] = [],
    ) -> None:
        """Initialize abstract feature applicator.

        Args:
            horizons: list of horizons in hours
            feature_names:  List of requested features
            feature_modules: List of modules from which FeatureAdders should be loaded.

        """
        if not isinstance(horizons, str) and type(horizons) is not list and not None:
            raise ValueError("horizons must be added as a list")

        self.feature_names = feature_names
        self.horizons = horizons
        self.features_adder = adders_from_modules(feature_modules)
        self.features_dispatcher = FeatureDispatcher(self.features_adder)

    @abstractmethod
    def add_features(
        self, df: pd.DataFrame, pj: PredictionJobDataClass = None
    ) -> pd.DataFrame:
        """Adds features to an input DataFrame.

        Args:
            df: DataFrame with input data to which the features have to be added
            pj: (Optional) A prediction job that is needed for location dependent features,
                if not specified a default location is used

        Returns:
            Dataframe with added features.

        """


class TrainFeatureApplicator(AbstractFeatureApplicator):
    """Feature applicator for use during training."""

    def add_features(
        self,
        df: pd.DataFrame,
        pj: PredictionJobDataClass = None,
        latency_config: dict = None,
    ) -> pd.DataFrame:
        """Adds features to an input DataFrame.

        This method is implemented specifically for a model train pipeline. For larger
        horzions data is invalidated as when they are not available.

        For example:
            For horzion 24 hours the feature T-720min is not added as the load
            720 minutes ago is not available 24 hours in advance. In case of a horizon
            0.25 hours this feature is added as in this case the feature is available.

        Args:
            df: Input data to which the features will be added.
            pj: (Optional) A prediction job that is needed for location dependent features,
                if not specified a default location is used
            latency_config: (Optional) Invalidate certain features that are not
                available for a specific horizon due to data latency. Defaults to
                ``{"APX": 24}``.

        Returns:
            Input DataFrame with an extra column for every added feature and sorted on the datetime index.

        """
        # If pj is none add empty dict
        if pj is None:
            pj = {}

        if latency_config is None:
            latency_config = LATENCY_CONFIG

        # Set default horizons if none are provided
        if self.horizons is None:
            self.horizons = [0.25, 24]

        # Pre define output variables
        result = pd.DataFrame()

        if isinstance(self.horizons, str):
            # copy the custom horizon into the horizon column
            res = df.copy(deep=True)
            res["horizon"] = res[self.horizons]
            result = pd.concat([result, res])
        else:
            # Loop over horizons and add corresponding features
            for horizon in self.horizons:
                # Deep copy of df is important, because we want a fresh start every iteration!
                res = apply_features(
                    df.copy(deep=True),
                    horizon=horizon,
                    pj=pj,
                    feature_names=self.feature_names,
                )
                res["horizon"] = horizon
                result = pd.concat([result, res])

        # Add custom features with the dispatcher
        result = self.features_dispatcher.apply_features(
            result, feature_names=self.feature_names
        )

        # IMPORTANT: sort index to prevent errors when slicing on the (datetime) index
        # if we don't sort, the duplicated indexes (one per horizon) have large gaps
        # and slicing will give an exception.
        result = result.sort_index(axis=0)

        # Invalidate features that are not available for a specific horizon due to data
        # latency
        for feature, time in latency_config.items():
            result.loc[result["horizon"] > time, feature] = np.nan

        # NOTE this is required since apply_features could add additional features
        if self.feature_names is not None:
            # Add horizon to requested features else it is removed, and if needed the proloaf feature (historic_load)
            if pj.get("model") == "proloaf":
                features = self.feature_names + ["historic_load"] + ["horizon"]
            else:
                features = self.feature_names + ["horizon"]
            result = remove_non_requested_feature_columns(result, features)

        # Sort all features except for the (first) load and (last) horizon columns
        return enforce_feature_order(result)


class OperationalPredictFeatureApplicator(AbstractFeatureApplicator):
    """Feature applicator for use in operational forecasts."""

    def add_features(
        self, df: pd.DataFrame, pj: PredictionJobDataClass = None
    ) -> pd.DataFrame:
        """Adds features to an input DataFrame.

        This method is implemented specifically for an operational prediction pipeline and will add every available feature.

        Args:
            df: DataFrame with input data to which the features have to be added
            pj: (Optional) A prediction job that is needed for location dependent features,
                if not specified a default location is used
        Returns:
           Input DataFrame with an extra column for every added feature.

        """
        # If pj is none add empty dict
        if pj is None:
            pj = {}

        num_horizons = len(self.horizons)
        if num_horizons != 1:
            raise ValueError(f"Expected one horizon, got {num_horizons}")

        # Add core features
        df = apply_features(
            df, feature_names=self.feature_names, horizon=self.horizons[0], pj=pj
        )
        # Add custom features with the dispatcher
        df = self.features_dispatcher.apply_features(
            df, feature_names=self.feature_names
        )

        df = add_missing_feature_columns(df, self.feature_names)

        # NOTE this is required since apply_features could add additional features
        if self.feature_names is not None:
            df = remove_non_requested_feature_columns(df, self.feature_names)

        return enforce_feature_order(df)



# SPDX-FileCopyrightText: 2017-2023 Contributors to the OpenSTEF project <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0
import traceback
from typing import Callable

import structlog

from openstef.exceptions import PredictionJobException
from openstef.monitoring.performance_meter import PerformanceMeter
from openstef.monitoring.teams import post_teams


class TaskContext:
    """A context manager that can be used to run tasks with.

    Should be used as:

    .. code-block:: py

        with TaskContext("my_task_name") as context:
            pass

    Args:
        name: Name of the task
        config: Configuration object, can be found in openSTEF-dbc
        database: Database object, can be found in openSTEF-dbc
        suppress_exceptions: If set to False the context
            manager will pass any raised exception on. Defaults to False.
        post_teams_on_exception: If set to True the context
            manager will automatically post a message to teams when an
            exception is encountered. Defaults to True.
        on_exception : Callback, will be called
            when an exception is raised. Callable gets exc_type, exc_info,
            stack_info as arguments.
        on_successful: Callback, will be called
            everytime if the task is successful (no exception is raised).
        on_end: Callback, will be called if the task is
            completed. Callable gets a bool indicating success as argument.

    """

    def __init__(
        self,
        name: str,
        config: object,
        database: object,
        suppress_exceptions: bool = False,
        post_teams_on_exception: bool = True,
        on_exception: Callable = None,
        on_successful: Callable = None,
        on_end: Callable = None,
    ):
        """Initialize context manager."""
        self.name = name
        self.suppress_exceptions = suppress_exceptions
        self.post_teams_on_exception = post_teams_on_exception
        self.on_exception = on_exception
        self.on_successful = on_successful
        self.on_end = on_end
        self.config = config
        self.database = database

    def __enter__(self):
        self.logger = structlog.get_logger(__name__).bind(task=self.name)

        self.perf_meter = PerformanceMeter(self.logger)

        self.perf_meter.start_level("task", self.name)

        return self

    def __exit__(self, exc_type: type, exc_info: str, stack_info: str) -> bool:
        self.successful = exc_type is None

        if self.successful:
            if self.on_successful is not None:
                self.on_successful()
        else:
            self.logger.error(
                "An error has occured",
                exc_info=exc_info,
                stack_info=stack_info,
            )
            if self.on_exception is not None:
                self.on_exception(exc_type, exc_info, stack_info)

            if self.post_teams_on_exception:
                self._send_teams_message(exc_type, exc_info, stack_info)

        if self.on_end is not None:
            self.on_end(self.successful)

        self.perf_meter.complete_level(self.successful)

        # Returning true stops the propagation of the exception
        return self.suppress_exceptions

    def _send_teams_message(
        self, exc_type: type, exc_info: str, stack_info: str
    ) -> None:
        stack_text = "".join(traceback.format_exception(exc_type, exc_info, stack_info))
        # The teams webhook behaves weird. Using the correct order \r\n does NOT work
        # The 4 spaces forces teams to format a monospaced message for some reason
        stack_text = stack_text.replace("\n", "\n\r    ")
        msg = {
            "title": f"Task '{self.name}' raised an exception",
            "color": "#a54f4f",
            "sections": [],
        }
        if exc_type is PredictionJobException:
            metrics = exc_info.metrics
            msg["sections"].append(
                {
                    "facts": [
                        ("Number of pjs in this task", metrics["num_jobs"]),
                        ("Number of pjs started", metrics["jobs_started"]),
                        ("Number of pjs successful", metrics["jobs_successful"]),
                        ("Number of pjs unsuccessful", metrics["jobs_unsuccessful"]),
                    ],
                }
            )
            string_pids_unsuccessful = ", ".join(
                [f"{pid:d}" for pid in metrics["pids_unsuccessful"]]
            )
            msg["sections"].append(
                {
                    "facts": [
                        ("Unsuccessful pid(s)", string_pids_unsuccessful),
                    ]
                }
            )
            # Add details of exceptions and pids
            string_pids_exceptions = "\n".join(
                [f"{key}:{value}\n" for key, value in metrics["exceptions"].items()]
            )
            msg["sections"].append(
                {
                    "facts": [
                        ("Exceptions: pid(s)", string_pids_exceptions),
                    ]
                }
            )

        msg["sections"].append(
            {
                "text": f"    {stack_text}",
            }
        )

        post_teams(
            msg, url=self.config.teams_monitoring_url, proxies=self.config.proxies
        )



# SPDX-FileCopyrightText: 2017-2023 Alliander N.V. <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0
"""This module defines the custom regressor."""
import inspect
from abc import abstractmethod
from importlib import import_module
from typing import Type

import pandas as pd

from openstef.model.objective import (
    EVAL_METRIC,
    TEST_FRACTION,
    VALIDATION_FRACTION,
    RegressorObjective,
)
from openstef.model.regressors.regressor import OpenstfRegressor


class CustomOpenstfRegressor(OpenstfRegressor):
    """A custom regressor allows to load any custom model that is not included with openSTEF."""

    @staticmethod
    @abstractmethod
    def valid_kwargs() -> list[str]:
        ...

    @classmethod
    @abstractmethod
    def objective(self) -> Type[RegressorObjective]:
        ...


def load_custom_model(custom_model_path) -> CustomOpenstfRegressor:
    """Load the external custom model."""
    path_elements = custom_model_path.split(".")
    module_path = ".".join(path_elements[:-1])
    module = import_module(module_path)
    model_name = path_elements[-1]
    model_class = getattr(module, model_name)

    if (
        not inspect.isclass(model_class)
        or inspect.isabstract(model_class)
        or not issubclass(model_class, CustomOpenstfRegressor)
    ):
        raise ValueError(
            f"The path {custom_model_path!r} does not correspond to a concrete"
            " CustomOpenstfRegressor subclass"
        )

    return model_class


def is_custom_type(model_type):
    return isinstance(model_type, str) and "." in model_type


def create_custom_objective(
    custom_model_path,
):
    model_class = load_custom_model(custom_model_path)
    return model_class.objective()



# SPDX-FileCopyrightText: 2017-2023 Alliander N.V. <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0
import structlog

from abc import ABC, abstractmethod
from typing import Optional

import pandas as pd
from datetime import timedelta
from openstef.data_classes.model_specifications import ModelSpecificationDataClass
from openstef.data_classes.prediction_job import PredictionJobDataClass
from openstef.model.regressors.regressor import OpenstfRegressor
from openstef.feature_engineering.feature_applicator import (
    TrainFeatureApplicator,
    OperationalPredictFeatureApplicator,
)
from openstef.feature_engineering.general import (
    enforce_feature_order,
    remove_non_requested_feature_columns,
)
from openstef.pipeline.utils import generate_forecast_datetime_range


class AbstractDataPreparation(ABC):
    def __init__(
        self,
        pj: PredictionJobDataClass,
        model_specs: ModelSpecificationDataClass,
        model: Optional[OpenstfRegressor] = None,
        horizons: Optional[list[float]] = None,
    ) -> None:
        super().__init__()
        self.pj = pj
        self.model_specs = model_specs
        self.model = model
        self.horizons = horizons

    @abstractmethod
    def prepare_train_data(self, data: pd.DataFrame) -> pd.DataFrame:
        pass

    @abstractmethod
    def prepare_forecast_data(
        self, data: pd.DataFrame
    ) -> tuple[pd.DataFrame, pd.DataFrame]:
        pass

    def check_model(self):
        if self.model is None:
            raise ValueError(
                "If no model has been provided to the data prep class, it cannot perform preparation for forecast task!"
            )


class LegacyDataPreparation(AbstractDataPreparation):
    def prepare_train_data(self, data: pd.DataFrame) -> pd.DataFrame:
        if self.horizons:
            horizons = self.horizons
        else:
            horizons = self.pj.resolution_minutes

        features_applicator = TrainFeatureApplicator(
            horizons=horizons,
            feature_names=self.model_specs.feature_names,
            feature_modules=self.model_specs.feature_modules,
        )
        return features_applicator.add_features(data, pj=self.pj)

    def prepare_forecast_data(self, data: pd.DataFrame) -> pd.DataFrame:
        self.check_model()

        features_applicator = OperationalPredictFeatureApplicator(
            horizons=[self.pj["resolution_minutes"] / 60.0],
            feature_names=self.model.feature_names,
            feature_modules=self.model_specs.feature_modules,
        )
        data_with_features = features_applicator.add_features(data)

        # Prep forecast input by selecting only the forecast datetime interval (this is much smaller than the input range)
        # Also drop the load column
        forecast_start, forecast_end = generate_forecast_datetime_range(
            data_with_features
        )
        forecast_input_data = data_with_features[forecast_start:forecast_end].drop(
            columns="load"
        )

        return forecast_input_data, data_with_features


class ARDataPreparation(AbstractDataPreparation):
    def __init__(
        self,
        pj: PredictionJobDataClass,
        model_specs: ModelSpecificationDataClass,
        model: Optional[OpenstfRegressor] = None,
        horizons: Optional[list[float]] = None,
        historical_depth: Optional[int] = None,
    ) -> None:
        super().__init__(pj, model_specs, model, horizons)
        self.historical_depth = historical_depth

    def prepare_train_data(self, data: pd.DataFrame) -> pd.DataFrame:
        # Add dummy horizon column
        data["horizon"] = 0

        # remove non requested feature
        features = self.model_specs.feature_names + ["horizon"]
        result = remove_non_requested_feature_columns(data, features)

        # Sort all features except for the (first) load and (last) horizon columns
        result = result[["load"] + [c for c in result.columns if c != "load"]]
        result = result.sort_index()
        result = enforce_feature_order(result)

        result = result[result.iloc[:, 0].notna()]
        return result

    def prepare_forecast_data(
        self, data: pd.DataFrame
    ) -> tuple[pd.DataFrame, pd.DataFrame]:
        logger = structlog.get_logger(__name__)
        self.check_model()
        # Prep forecast input by selecting only the forecast datetime interval (this is much smaller than the input range)
        # Also drop the load column
        data = data[["load"] + self.model.feature_names]
        forecast_start, forecast_end = generate_forecast_datetime_range(data)
        forecast_input_data = data[forecast_start:forecast_end].drop(columns="load")

        historical_start = None
        if self.historical_depth:
            historical_start = forecast_start - self.historical_depth * timedelta(
                minutes=self.pj.resolution_minutes
            )
        past_data = data[historical_start:forecast_start].iloc[:-1]
        self.model.update_historic_data(
            past_data.drop(columns="load"), past_data["load"]
        )
        logger.info(
            "Watch-out side effect on the model performed in the feature builder to update the historical data."
        )

        data[self.model.feature_importance_dataframe.index.tolist()] = 0
        return forecast_input_data, data



# SPDX-FileCopyrightText: 2017-2023 Contributors to the OpenSTEF project <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0

import random
import sys
from collections import defaultdict
from itertools import chain
from typing import Callable

from openstef.data_classes.prediction_job import PredictionJobDataClass
from openstef.exceptions import PredictionJobException
from openstef.tasks.utils.dependencies import find_groups, has_dependencies
from openstef.tasks.utils.taskcontext import TaskContext


class PredictionJobLoop:
    """Convenience objects that maps a function over prediction jobs.

    Default behaviour is to automatically get prediction jobs from the
    database. Any keyword argument passed will be directed to the getting
    function. If another set of prediction jobs is desired, manually pass
    them using the prediction_jobs argument.

    Tip: For debugging a specific PID, use debug_pid=specific_pid

    Args:
        context: The context to run this loop in.
        stop_on_exception: Whether to break out of the
            loop when an exception is raised. Defaults to False.
        random_order: Whether to randomize the order of the
            prediction jobs. Defaults to True. Does not apply to manually
            passed prediction jobs.
        on_exception_callback: Callback, will be called
            everytime an exception is raised. Callable gets the pj and
            exception raised as arguments
        on_successful_callback: Callback, will be called
            everytime an iteration is successful (no exception is raised).
            Callable gets the pj as argument.
        on_end_callback: Callback, will be called everytime an
            iteration is completed. Callable gets the pj and and bool
            indicating success as argument.
        prediction_jobs: Manually pass a list of
            prediction jobs that will be looped over. If set to None, will fetch
            prediction jobs from database based on pj_kwargs
        debug_pid: enter a specific pid for debugging.
            If not None, the prediction job loop will only look at this pid
        **pj_kwargs: Any other kwargs willed will be directed to the
            prediction job getting function.

    """

    def __init__(
        self,
        context: TaskContext,
        stop_on_exception: bool = False,
        random_order: bool = True,
        on_exception_callback: Callable = None,
        on_successful_callback: Callable = None,
        on_end_callback: Callable = None,
        prediction_jobs: list[PredictionJobDataClass] = None,
        debug_pid: int = None,
        **pj_kwargs
    ):
        """Initialzie prediction job loop."""
        self.context = context
        self.stop_on_exception = stop_on_exception
        self.random_order = random_order
        self.on_exception_callback = on_exception_callback
        self.on_successful_callback = on_successful_callback
        self.on_end_callback = on_end_callback
        self.pj_kwargs = pj_kwargs
        self.debug_pid = debug_pid

        if prediction_jobs is None:
            self.prediction_jobs = self._get_prediction_jobs()
        else:
            self.prediction_jobs = prediction_jobs

        if has_dependencies(self.prediction_jobs):
            # Groups so that prediction jobs in each group
            # depend only on the pj in the previous groups
            _, pj_groups = find_groups(self.prediction_jobs, self.random_order)
            # Concatenate groups
            self.prediction_jobs = list(chain(*pj_groups))
        else:
            if self.random_order:
                random.shuffle(self.prediction_jobs)

    def _get_prediction_jobs(self):
        """Fetches prediction jobs from database."""
        self.context.logger.info(
            "Querying prediction jobs from database", **self.pj_kwargs
        )
        if self.debug_pid is None:
            prediction_jobs = self.context.database.get_prediction_jobs(
                **self.pj_kwargs
            )
        else:
            # retrieve prediction_job for single pid - useful for debugging
            prediction_jobs = [
                self.context.database.get_prediction_job(
                    pid=self.debug_pid, **self.pj_kwargs
                )
            ]

        return prediction_jobs

    def map(self, function: Callable, *args, **kwargs):
        """Maps the passed function over all prediction jobs.

        Args:
            function: The function that will be applied to each prediction
                job separately.
            *args: Any other arguments or passed to the function.
            **kwargs: All keyword arguments are passed to the function. This method is
                highly prefered over using args, since kwargs will be automatically
                logged.

        Raises:
            PredictionJobException: This exception will be raised if one or more
                prediction jobs raised an exception during the loop.

        """
        pids_successful = []
        pids_unsuccessful = []
        pids_unsuccessful_dict = defaultdict(list)
        last_job_exception = None

        num_jobs = len(self.prediction_jobs)
        self.context.perf_meter.checkpoint("pre-loop")

        # loop over prediction jobs
        for i, prediction_job in enumerate(self.prediction_jobs):
            successful = False

            self.context.logger = self.context.logger.bind(
                prediction_id=prediction_job["id"],
                prediction_name=prediction_job["name"],
            )

            self.context.perf_meter.start_level(
                "iteration", i, num_jobs=num_jobs, pid=prediction_job["id"], **kwargs
            )

            try:
                function(prediction_job, *args, **kwargs)

                pids_successful.append(prediction_job["id"])
                successful = True

                self._handle_successful_iteration(prediction_job)

            except Exception as exception:
                pids_unsuccessful.append(prediction_job["id"])
                pids_unsuccessful_dict[str(exception)].append(prediction_job["id"])
                last_job_exception = exception

                self._handle_exception_during_iteration(prediction_job, exception)

                if self.stop_on_exception:
                    break
            finally:
                self._handle_finished_last_iteration(prediction_job, successful)

            self.context.logger = self.context.logger.unbind("prediction_id")

        jobs_successful = len(pids_successful)
        jobs_unsuccessful = len(pids_unsuccessful)
        jobs_started = jobs_successful + jobs_unsuccessful

        # This log is for human readable logging
        self.context.perf_meter.checkpoint(
            "loop",
            num_jobs=num_jobs,
            jobs_started=jobs_started,
            jobs_successful=jobs_successful,
            jobs_unsuccessful=jobs_unsuccessful,
            successful=int(jobs_unsuccessful > 0),
        )

        if jobs_unsuccessful > 0:
            metrics = {
                "num_jobs": num_jobs,
                "pids_successful": pids_successful,
                "pids_unsuccessful": pids_unsuccessful,
                "exceptions": pids_unsuccessful_dict,
                "jobs_successful": jobs_successful,
                "jobs_unsuccessful": jobs_unsuccessful,
                "jobs_started": jobs_started,
            }

            raise PredictionJobException(metrics) from last_job_exception

    def _handle_successful_iteration(self, prediction_job):
        if self.on_successful_callback is not None:
            try:
                self.on_successful_callback(prediction_job)
            except Exception:
                _, exc_info, stack_info = sys.exc_info()
                self.context.logger.error(
                    "An exception occured when executing the on_successful_callback    "
                    "                    callback function for this iteration",
                    exc_info=exc_info,
                    stack_info=stack_info,
                )

    def _handle_exception_during_iteration(self, prediction_job, e):
        _, exc_info, stack_info = sys.exc_info()
        self.context.logger.error(
            "An exception occured during this iteration",
            exc_info=exc_info,
            stack_info=stack_info,
        )

        if self.on_exception_callback is None:
            return

        try:
            self.on_exception_callback(prediction_job, e)
        except Exception:
            _, exc_info, stack_info = sys.exc_info()
            self.context.logger.error(
                "An exception occured when executing the on_exception_callback "
                "callback function for this iteration",
                exc_info=exc_info,
                stack_info=stack_info,
            )

    def _handle_finished_last_iteration(self, prediction_job, successful):

        self.context.perf_meter.complete_level(successful)

        if self.on_end_callback is not None:
            try:
                self.on_end_callback(prediction_job, successful)
            except Exception:
                _, exc_info, stack_info = sys.exc_info()
                self.context.logger.error(
                    "An exception occured when executing the on_end_callback callback  "
                    "                  function for this iteration",
                    exc_info=exc_info,
                    stack_info=stack_info,
                )



# SPDX-FileCopyrightText: 2017-2023 Contributors to the OpenSTEF project <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0
"""This module contains all functions for generating figures."""
import base64
from typing import Union

import pandas as pd
import plotly.express as px
import plotly.graph_objects as go


def plot_feature_importance(feature_importance: pd.DataFrame) -> go.Figure:
    """Created a treemap plot based on feature importance and weights.

    Args:
        feature_importance: A DataFrame describing the
            feature importances and weights of the trained model.

    Returns:
        A treemap of the features.

    """
    feature_importance["parent"] = "Feature importance"

    return go.Figure(
        go.Treemap(
            labels=feature_importance.index,
            parents=feature_importance["parent"],
            values=feature_importance["gain"],
            customdata=feature_importance["weight"],
            marker=dict(colors=feature_importance["weight"], colorscale="greens"),
            hovertemplate=(
                "<b>%{label}</b><br>importance: %{value:.1%}"
                "<br>weight: %{customdata:.1%}<extra></extra>"
            ),
        ),
        layout={
            "margin": {
                "t": 0,
                "r": 0,
                "b": 0,
                "l": 0,
            }
        },
    )


def plot_data_series(
    data: Union[list[pd.DataFrame], list[pd.Series]],
    predict_data: Union[list[pd.DataFrame], list[pd.Series]] = None,
    horizon: int = 47,
    names: list[str] = None,
) -> go.Figure:
    """Plots passed data and optionally prediction data for specified horizon.

    Args:
        data: There are two options to use this function. Either
            pass a list of pandas.DataFrame where each dataframe contains a load
            column and a horizon column. Or pass a list of pandas.Series with
            unique indexing.
        predict_data: Similar to data, but
            for prediction data instead. When passing a list of pandas.DataFrame
            the column forecast should exist. Can be set to None.
        horizon: This function will select only data matching
            this horizon. Defaults to 47.
        names: The names that will be used in the
            legend of the plot. If None is passed, this will be build
            automatically based on the number of series passed.

    Returns:
        A line plot of each passed data series.

    """
    series_names = {
        1: ("series",),
        2: ("train", "validation"),
        3: ("train", "validation", "test"),
    }

    num_series = len(data)

    if names is None and num_series > 3:
        raise ValueError(
            "Cannot pass names=None when passing data with more than 3 series."
        )

    if names is None:
        names = series_names[num_series]

    if predict_data is None:
        # Check if horizon columns exists in the data
        if "horizon" in data[0]:
            # Filter data on given horizon
            filtered = []
            for series in data:
                mask = series["horizon"] == horizon
                filtered.append(series[mask]["load"])
        else:
            filtered = data

        return _plot_data(names, filtered)

    # Check if horizon columns exists in the data
    if "horizon" in data[0]:
        # Filter data on given horizon
        actuals = []
        predictions = []

        for series, predict_series in zip(data, predict_data):
            mask = series["horizon"] == horizon
            actuals.append(series[mask]["load"])
            predictions.append(predict_series[mask]["forecast"])
    else:
        actuals = data
        predictions = predict_data

    fig = _plot_data_and_predictions(names, actuals, predictions)
    fig.update_layout(
        title=f"Predictor in action for horizon: {horizon}",
    )

    return fig


def _plot_data(names: list[str], series: list[pd.Series]) -> go.Figure:
    """Create plot of data consisting of different splits.

    Note:
        Do not use this function directly, use plot_data_series instead.

    Args:
        names: Name of each seperate split.
        series: Each split as a seperate series.

    Returns:
        A line plot of each passed series.

    """
    # Build a combined DataFrame with all data.
    # This step is important to create forced NaNs to create gaps in the plot.
    combined = []
    for name, sequence in zip(names, series):
        combined.extend(
            [
                sequence.rename(name),
            ]
        )
    df_plot = pd.concat(combined, axis=1)

    fig = go.Figure()

    # Add a trace for every data series
    for i, name in enumerate(names):
        fig.add_trace(
            go.Scatter(
                x=df_plot.index,
                y=df_plot[name],
                name=name,
                line=dict(color=px.colors.qualitative.Set2[i]),
            )
        )

    fig.update_layout(yaxis_title="Load (MW)")

    return fig


def _plot_data_and_predictions(
    names: list[str], actuals: list[pd.Series], predictions: list[pd.Series]
) -> go.Figure:
    """Create plot of different data and prediction splits.

    Note:
        Do not use this function directly, use plot_data_series instead.

    Args:
        names: Name of each seperate split. The passed names will be suffixed with _actual and _predict for data and predictions respectively.
        actuals: Each data split as a seperate series.
        predictions: Each prediction split as a seperate series.

    Returns:
        A line plot of each passed series.

    """
    # Build a combined DataFrame with all data.
    # This step is important to create forced NaNs to create gaps in the plot.
    combined = []
    for name, actual, prediction in zip(names, actuals, predictions):
        combined.extend(
            [
                actual.rename(f"{name}_actual"),
                prediction.rename(f"{name}_predict"),
            ]
        )
    df_plot = pd.concat(combined, axis=1)

    fig = go.Figure()

    # Add a trace for every data series
    for i, name in enumerate(names):
        actual, predict = f"{name}_actual", f"{name}_predict"

        fig.add_trace(
            go.Scatter(
                x=df_plot.index,
                y=df_plot[actual],
                name=actual,
                line=dict(color=px.colors.qualitative.Set2[i]),
            )
        )
        fig.add_trace(
            go.Scatter(
                x=df_plot.index,
                y=df_plot[predict],
                name=predict,
                line=dict(dash="dot", color=px.colors.qualitative.Dark2[i]),
            )
        )

    fig.update_layout(yaxis_title="Load (MW)")

    return fig


def convert_to_base64_data_uri(path_in: str, path_out: str, content_type: str) -> None:
    """Read file, convert it to a data_uri, then writes the data_uri to file.

    Args:
        path_in: Path of the file that will be converted
        path_out: Path of the file containing the data uri
        content_type: Content type of the data uri according to
            (https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Content-Type).

    """
    with open(path_in, "rb") as f:
        encoded = base64.b64encode(f.read()).decode()

    data_uri = "data:{0};base64,{1}".format(content_type, encoded)

    with open(path_out, "wt") as f:
        f.write(data_uri)



# SPDX-FileCopyrightText: 2017-2023 Alliander N.V. <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0
"""This module provides functionality for defining custom feature adders."""
import inspect
import re
from abc import ABC, abstractmethod
from collections import Counter, namedtuple
from importlib import import_module
from typing import Optional, Sequence

import pandas as pd

ParsedFeature = namedtuple("ParsedFeature", ["name", "params"])


class FeatureAdder(ABC):
    """Abstract class that implement the FeatureAdder interface.

    It is the basic block that handles the logic for computing the specific feature and the syntactic sugar to load
    properly the feature adder according to the feature name.

    """

    @property
    @abstractmethod
    def _regex(self) -> str:
        pass

    @property
    @abstractmethod
    def name(self) -> str:
        """Name of the FeatureAdder."""

    @abstractmethod
    def required_features(self, feature_names) -> list[str]:
        """List of features that are required to calculate this feature."""

    def __hash__(self):
        """Genearate hash of the name of this feature."""
        return hash(self.name)

    def parse_feature_name(self, feature_name: str) -> Optional[dict[str, str]]:
        """Parse a feature name.

        If the feature name is taken in charge by the feature adder, the method returns
        a dictionnary with the potentially parsed parameters contained the feature name. In the
        case the feature name does not contain parameters an empty dictionary is returned.
        Otherwise the method returns None.

        Args:
            feature_name (str): The feature name, this may contain parameter informations.

        Returns:
            Optional[dict[str, Any]]: The parsed parameters. If the feature name is recognized but has no parameters
            an empty dictionnary is returned. If the feature name is not recognized, None is
            returned.

        """
        reg = self._regex
        match = re.match(reg, feature_name)
        return None if match is None else match.groupdict()

    @abstractmethod
    def apply_features(
        self, df: pd.DataFrame, parsed_feature_names: Sequence[ParsedFeature]
    ) -> pd.DataFrame:
        """Apply or add the features to the input dataframe."""

    def __repr__(self):
        """Represent as string."""
        return "%s(<%s>)" % (self.__class__.__name__, self.name)


class FeatureDispatcher:
    """Orchestrator of the feature adders.

    It scans the feature_names to assign to each feature the proper feature adder and launch the effective computing of
    the features.

    """

    def __init__(self, feature_adders: Sequence[FeatureAdder]):
        """Initialize feature dispatcher."""
        self.feature_adders = list(feature_adders)
        self._check_feature_adder_names_unicity()

    def _check_feature_adder_names_unicity(self):
        names = Counter(adder.name for adder in self.feature_adders)
        duplicated_names = []
        for name, count in names.items():
            if count > 1:
                duplicated_names.append(name)

        if len(duplicated_names) > 0:
            raise RuntimeError(
                "There is at least one duplicated feature adder name: %s"
                % duplicated_names
            )

    def dispatch_features(
        self, feature_names: list[str]
    ) -> dict[FeatureAdder, list[ParsedFeature]]:
        """Dispatch features.

        Args:
            feature_names: The names of the features to be dispatched.

        Returns:
            Dictionary with parsed features.

        """
        recognized_features = set()
        dispatched_features = {}

        for feature_name in feature_names:
            for adder_obj in self.feature_adders:
                parsed_params = adder_obj.parse_feature_name(feature_name)
                if parsed_params is not None:
                    if feature_name in recognized_features:
                        raise RuntimeError(
                            "Ambiguous feature adder set detected. The feature name"
                            " '%s' is recognised by more than 1 feature adder"
                            % feature_names
                        )
                    recognized_features.add(feature_name)
                    features = dispatched_features.setdefault(adder_obj, [])
                    features.append(ParsedFeature(feature_name, parsed_params))

        return dispatched_features

    def apply_features(
        self, df: pd.DataFrame, feature_names: list[str]
    ) -> pd.DataFrame:
        """Applies features to the input DataFrame.

        Args:
            df: DataFrame to which the features have to be added.
            feature_names: Names of the features.

        Returns:
            DataFrame with the added features.

        """
        if feature_names is None:
            return df
        dispatched_features = self.dispatch_features(feature_names)

        applied_features = set()
        applied_features_num = 0

        while True:
            for adder, parsed_features in dispatched_features.items():
                parsed_feature_names = [pf.name for pf in parsed_features]
                required_features = adder.required_features(parsed_feature_names)

                if len(set(required_features) - set(df.columns)) == 0:
                    df = adder.apply_features(df, parsed_features)
                    applied_features |= set(parsed_feature_names)

            if (
                len(applied_features) == applied_features_num
            ):  # No new feature was treated
                break

            applied_features_num = len(applied_features)

        return df


def adders_from_module(module_name: str) -> list[FeatureAdder]:
    """Load all FeatureAdders classes on the fly from the module.

    Args:
        module_name: The name of the module from which to import.

    Returns:
        A list with all loaded FeatureAdders.

    """
    module = import_module(module_name)
    feature_adders = []

    for element_name in dir(module):
        element = getattr(module, element_name)
        if (
            isinstance(element, type)
            and issubclass(element, FeatureAdder)
            and not inspect.isabstract(element)
        ):
            feature_adders.append(element())

    return feature_adders


def adders_from_modules(module_names: list[str]) -> list[FeatureAdder]:
    """Load all FeatureAdders classes on the fly from multiple modules.

    Args:
        module_names: A list with names of the modules from which to import.

    Returns:
        A list with all loaded FeatureAdders.

    """
    return sum((adders_from_module(module_name) for module_name in module_names), [])



# SPDX-FileCopyrightText: 2017-2023 Contributors to the OpenSTEF project <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0
"""This module contains the CRON job that is periodically executed to retrain the prognosis models.

For this the folowing steps are caried out:
  1. Get historic training data (TDCV, Load, Weather and APX price data)
  2. Apply features
  3. Train and Test the new model
  4. Check if new model performs better than the old model
  5. Store the model if it performs better
  6. Send slack message to inform the users

Example:
    This module is meant to be called directly from a CRON job. A description of
    the CRON job can be found in the /k8s/CronJobs folder.
    Alternatively this code can be run directly by running::

        $ python model_train.py

"""
from datetime import datetime, timedelta
from pathlib import Path

from openstef.data_classes.prediction_job import PredictionJobDataClass

from openstef.enums import MLModelType, PipelineType
from openstef.exceptions import SkipSaveTrainingForecasts
from openstef.pipeline.train_model import train_model_pipeline
from openstef.tasks.utils.predictionjobloop import PredictionJobLoop
from openstef.tasks.utils.taskcontext import TaskContext

TRAINING_PERIOD_DAYS: int = 120
DEFAULT_CHECK_MODEL_AGE: bool = True


def train_model_task(
    pj: PredictionJobDataClass,
    context: TaskContext,
    check_old_model_age: bool = DEFAULT_CHECK_MODEL_AGE,
    datetime_start: datetime = None,
    datetime_end: datetime = None,
) -> None:
    """Train model task.

    Top level task that trains a new model and makes sure the beast available model is
    stored. On this task level all database and context manager dependencies are resolved.

    Expected prediction job keys:  "id", "model", "lat", "lon", "name"

    Args:
        pj: Prediction job
        context: Contect object that holds a config manager and a
            database connection.
        check_old_model_age: check if model is too young to be retrained
        datetime_start: Start
        datetime_end: End

    """
    # Check pipeline types
    if PipelineType.TRAIN not in pj.pipelines_to_run:
        context.logger.info(
            "Skip this PredictionJob because train pipeline is not specified in the pj."
        )
        return

    # Get the paths for storing model and reports from the config manager
    mlflow_tracking_uri = context.config.paths_mlflow_tracking_uri
    context.logger.debug(f"MLflow tracking uri: {mlflow_tracking_uri}")
    artifact_folder = context.config.paths_artifact_folder
    context.logger.debug(f"Artifact folder: {artifact_folder}")

    context.perf_meter.checkpoint("Added metadata to PredictionJob")

    # Define start and end of the training input data
    if datetime_end is None:
        datetime_end = datetime.utcnow()
    if datetime_start is None:
        datetime_start = datetime_end - timedelta(days=TRAINING_PERIOD_DAYS)

    # todo: See if we can check model age before getting the data
    # Get training input data from database
    input_data = context.database.get_model_input(
        pid=pj["id"],
        location=[pj["lat"], pj["lon"]],
        datetime_start=datetime_start,
        datetime_end=datetime_end,
    )

    context.perf_meter.checkpoint("Retrieved timeseries input")

    # Excecute the model training pipeline
    try:
        data_sets = train_model_pipeline(
            pj,
            input_data,
            check_old_model_age=check_old_model_age,
            mlflow_tracking_uri=mlflow_tracking_uri,
            artifact_folder=artifact_folder,
        )

        if pj.save_train_forecasts:
            if data_sets is None:
                raise RuntimeError("Forecasts were not retrieved")
            if not hasattr(context.database, "write_train_forecasts"):
                raise RuntimeError(
                    "Database connector does dot support 'write_train_forecasts' while "
                    "'save_train_forecasts option was activated.'"
                )
            context.database.write_train_forecasts(pj, data_sets)
            context.logger.debug(f"Saved Forecasts from trained model on datasets")
    except SkipSaveTrainingForecasts:
        context.logger.debug(f"Skip saving forecasts")

    context.perf_meter.checkpoint("Model trained")


def main(model_type=None, config=None, database=None):
    if database is None or config is None:
        raise RuntimeError(
            "Please specifiy a config object and/or database connection object. These"
            " can be found in the openstef-dbc package."
        )

    if model_type is None:
        model_type = [ml.value for ml in MLModelType]

    taskname = Path(__file__).name.replace(".py", "")
    datetime_now = datetime.utcnow()
    with TaskContext(taskname, config, database) as context:
        PredictionJobLoop(context, model_type=model_type).map(
            train_model_task, context, datetime_end=datetime_now
        )


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: 2017-2023 Contributors to the OpenSTEF project <korte.termijn.prognoses@alliander.com>
#
# SPDX-License-Identifier: MPL-2.0
"""Specifies the dataclass for model specifications."""
from typing import Optional, Union

from pydantic import BaseModel


class ModelSpecificationDataClass(BaseModel):
    """Holds all information regarding the training procces of a specific model."""

    id: Union[int, str]
    hyper_params: Optional[dict] = {}
    """Hyperparameters that should be used during training."""
    feature_names: Optional[list] = None
    """Features that should be used during training."""
    feature_modules: Optional[list] = []
    """Feature modules that should be used during training."""

    def __getitem__(self, item: str) -> any:
        """Allows us to use subscription to get the items from the object."""
        return getattr(self, item)

    def __setitem__(self, key: str, value: any) -> None:
        """Allows us to use subscription to set the items in the object."""
        if hasattr(self, key):
            self.__dict__[key] = value
        else:
            raise AttributeError(f"{key} not an attribute of model specifications.")



# SPDX-FileCopyrightText: 2017-2023 Contributors to the OpenSTEF project <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0
"""Tracy checks the mysql todolist and tries her best to execute the functions with desired inputs.

This scripts works as follows:
  1. Checks the mysql table 'todolist' for jobs (which are not already in progress and
    which are not already failed)
  2. Set all newly acquired jobs to 'in progress'
For each job:
  3. Convert input arguments to a dict with 'args' and 'kwargs'
  4. Interpret the given function and arguments
  5. Execute the job
  6. Post result to Teams
  7. Remove job from mysql table
If job fails, set in progress to 2
All functions that tracy is able to execute need to be imported and defined in the
available_functions.

Example:
    This module is meant to be called directly from a CRON job.

    Alternatively this code can be run directly by running::
        $ python run_tracy.py

"""
# sql to create the Tracy jobs table (todolist)

# CREATE TABLE IF NOT EXISTS `tst_icarus`.`todolist` (
# `id` INT NOT NULL AUTO_INCREMENT ,
# `created` DATETIME NOT NULL DEFAULT CURRENT_TIMESTAMP ,
# `function` VARCHAR(200) NOT NULL ,
# `args` VARCHAR(200) NOT NULL ,
# `inprogress` BOOLEAN NULL DEFAULT NULL ,
# PRIMARY KEY (`id`), UNIQUE `id` (`id`))
# ENGINE = InnoDB;
from pathlib import Path

from openstef.data_classes.prediction_job import PredictionJobDataClass
from openstef.enums import TracyJobResult
from openstef.monitoring import teams
from openstef.tasks.optimize_hyperparameters import optimize_hyperparameters_task
from openstef.tasks.train_model import train_model_task
from openstef.tasks.utils.taskcontext import TaskContext


def run_tracy(context: TaskContext) -> None:
    # Get all Tracy jobs
    tracy_jobs = context.database.ktp_api.get_all_tracy_jobs(inprogress=0)
    num_jobs = len(tracy_jobs)

    if num_jobs == 0:
        context.logger.warning(f"Number of tracy jobs is {num_jobs}, exit task")
        return

    context.logger.info("Start processing Tracy jobs", num_jobs=num_jobs)

    for i, job in enumerate(tracy_jobs):

        # get a new logger with bound job
        logger = context.logger.bind(job=job)
        logger.info("Process job", job_counter=i, total_jobs=num_jobs)

        # Set all retrieved items of the todolist to inprogress
        job["inprogress"] = 1
        context.database.ktp_api.update_tracy_job(job)

        pid = int(job["args"])
        pj = context.database.get_prediction_job(pid)
        result, exc = run_tracy_job(job, pj, context)
        # job processing was succefull
        if result is TracyJobResult.SUCCESS:
            logger.info("Succesfully processed Tracy job")
            # Delete job when succesfull
            context.database.ktp_api.delete_tracy_job(job)
            logger.info("Delete Tracy job")

        # job was unknown
        elif result is TracyJobResult.UNKNOWN:
            logger.error(f"Unkown Tracy job {job['function']}")

        # job processing failed / raised an exception
        elif result is TracyJobResult.FAILED:
            job["inprogress"] = 2
            context.database.ktp_api.update_tracy_job(job)
            msg = "Exception occured while processing Tracy job"
            logger.error(msg, exc_info=exc)
            teams.post_teams(teams.format_message(title=msg, params=job))

    context.logger.info("Finished processing all Tracy jobs - Tracy out!")


def run_tracy_job(
    job: dict, pj: PredictionJobDataClass, context: TaskContext
) -> TracyJobResult:
    """Run tracy job.

    Args:
        job: Tracy jon
        pj: Prediction job
        context: Task context.

    Returns:
        Result of the Tracy job.

    """
    # Try to process Tracy job
    try:
        # If train model job (TODO remove old name when jobs are done)
        if job["function"] in ["train_model", "train_specific_model"]:
            train_model_task(pj, context, check_old_model_age=False)

        # If optimize hyperparameters job (TODO remove old name when jobs are done)
        elif job["function"] in [
            "optimize_hyperparameters",
            "optimize_hyperparameters_for_specific_pid",
        ]:
            optimize_hyperparameters_task(pj, context, check_hyper_param_age=False)

        # Else unknown job
        else:
            return TracyJobResult.UNKNOWN, None

        # job processing was succesfull
        return TracyJobResult.SUCCESS, None

    # Processing of Tracy job failed
    except Exception as e:
        return TracyJobResult.FAILED, e


def main(config=None, database=None):
    taskname = Path(__file__).name.replace(".py", "")

    if database is None or config is None:
        raise RuntimeError(
            "Please specifiy a config object and/or database connection object. These"
            " can be found in the openstef-dbc package."
        )

    with TaskContext(taskname, config, database) as context:
        run_tracy(context)


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: 2017-2023 Contributors to the OpenSTEF project <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0
import pandas as pd
import structlog

from openstef.data_classes.model_specifications import ModelSpecificationDataClass
from openstef.data_classes.prediction_job import PredictionJobDataClass
from openstef.feature_engineering.feature_applicator import (
    OperationalPredictFeatureApplicator,
)
from openstef.model.confidence_interval_applicator import ConfidenceIntervalApplicator
from openstef.model.fallback import generate_fallback
from openstef.model.regressors.regressor import OpenstfRegressor
from openstef.model.serializer import MLflowSerializer
from openstef.pipeline.utils import generate_forecast_datetime_range
from openstef.postprocessing.postprocessing import (
    add_prediction_job_properties_to_forecast,
)
from openstef.validation import validation


def create_forecast_pipeline(
    pj: PredictionJobDataClass,
    input_data: pd.DataFrame,
    mlflow_tracking_uri: str,
) -> pd.DataFrame:
    """Create forecast pipeline.

    This is the top-level pipeline which included loading the most recent model for
    the given prediction job.

    Expected prediction job keys: "id",

    Args:
        pj: Prediction job
        input_data: Training input data (without features)
        mlflow_tracking_uri: MlFlow tracking URI

    Returns:
        DataFrame with the forecast

    """
    prediction_model_pid = pj["id"]
    # Use the alternative forecast model if it's specify in the pj
    if pj.alternative_forecast_model_pid:
        prediction_model_pid = pj.alternative_forecast_model_pid

    # Load most recent model for the given pid
    model, model_specs = MLflowSerializer(
        mlflow_tracking_uri=mlflow_tracking_uri
    ).load_model(experiment_name=str(prediction_model_pid))
    return create_forecast_pipeline_core(pj, input_data, model, model_specs)


def create_forecast_pipeline_core(
    pj: PredictionJobDataClass,
    input_data: pd.DataFrame,
    model: OpenstfRegressor,
    model_specs: ModelSpecificationDataClass,
) -> pd.DataFrame:
    """Create forecast pipeline (core).

    Computes the forecasts and confidence intervals given a prediction job and input data.
    This pipeline has no database or persisitent storage dependencies.

    Expected prediction job keys: "resolution_minutes", "horizon_minutes", "id", "type",
        "name", "quantiles"

    Args:
        pj: Prediction job.
        input_data: Input data for the prediction.
        model: Model to use for this prediction.
        model_specs: Model specifications.

    Returns:
        Forecast

    """
    logger = structlog.get_logger(__name__)

    fallback_strategy = "extreme_day"  # this can later be expanded

    # Validate and clean data
    validated_data = validation.validate(pj["id"], input_data, pj["flatliner_treshold"])

    # Custom data prep or legacy behavior
    if pj.data_prep_class:
        data_prep_class, data_prep_args = pj.data_prep_class.load()
        forecast_input_data, data_with_features = data_prep_class(
            pj=pj,
            model_specs=model_specs,
            model=model,
            **data_prep_args,
        ).prepare_forecast_data(validated_data)
    else:
        # Add features
        data_with_features = OperationalPredictFeatureApplicator(
            horizons=[pj["resolution_minutes"] / 60.0],
            feature_names=model.feature_names,
            feature_modules=model_specs.feature_modules,
        ).add_features(validated_data)

        # Prep forecast input by selecting only the forecast datetime interval (this is much smaller than the input range)
        # Also drop the load column
        forecast_start, forecast_end = generate_forecast_datetime_range(
            data_with_features
        )
        forecast_input_data = data_with_features[forecast_start:forecast_end].drop(
            columns="load"
        )

    # Check if sufficient data is left after cleaning
    if not validation.is_data_sufficient(
        data_with_features,
        pj["completeness_treshold"],
        pj["minimal_table_length"],
        model,
    ):
        logger.warning(
            "Using fallback forecast",
            forecast_type="fallback",
            pid=pj["id"],
            fallback_strategy=fallback_strategy,
        )
        forecast = generate_fallback(data_with_features, input_data[["load"]])

    else:
        # Predict
        model_forecast = model.predict(forecast_input_data)
        forecast = pd.DataFrame(
            index=forecast_input_data.index, data={"forecast": model_forecast}
        )

    # Add confidence
    forecast = ConfidenceIntervalApplicator(
        model, forecast_input_data
    ).add_confidence_interval(forecast, pj)

    # Prepare for output
    forecast = add_prediction_job_properties_to_forecast(
        pj,
        forecast,
        algorithm_type=str(model.path),
    )

    return forecast



# SPDX-FileCopyrightText: 2017-2023 Alliander N.V. <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0
"""This module contains the linear regressor."""
import numpy as np
from sklearn.linear_model import LinearRegression
from sklearn.utils.validation import check_is_fitted

from openstef.model.metamodels.missing_values_handler import MissingValuesHandler
from openstef.model.regressors.regressor import OpenstfRegressor


class LinearRegressor(MissingValuesHandler):
    """Linear Regressor wrapped in the metamodel `MissingValuesHandler`.

    This regressor can handle missing values by imputation strategy.

    Args:
        missing_values : int, float, str, np.nan or None, default=np.nan
            The placeholder for the missing values. All occurrences of
            `missing_values` will be imputed. For pandas' dataframes with
            nullable integer dtypes with missing values, `missing_values`
            should be set to `np.nan`, since `pd.NA` will be converted to `np.nan`.

        imputation_strategy : str, default=None
            The imputation strategy.
            - If None no imputation is performed.
            - If "mean", then replace missing values using the mean along
            each column. Can only be used with numeric data.
            - If "median", then replace missing values using the median along
            each column. Can only be used with numeric data.
            - If "most_frequent", then replace missing using the most frequent
            value along each column. Can be used with strings or numeric data.
            If there is more than one such value, only the smallest is returned.
            - If "constant", then replace missing values with fill_value. Can be
            used with strings or numeric data.

        fill_value : str or numerical value, default=None
            When strategy == "constant", fill_value is used to replace all
            occurrences of missing_values.
            If left to the default, fill_value will be 0 when imputing numerical
            data and "missing_value" for strings or object data types.

    """

    def __init__(self, missing_values=np.nan, imputation_strategy=None, fill_value=0):
        """Initialize Linear regressor."""
        super().__init__(
            LinearRegression(),
            missing_values=missing_values,
            imputation_strategy=imputation_strategy,
            fill_value=fill_value,
        )


class LinearOpenstfRegressor(LinearRegressor, OpenstfRegressor):
    """Linear Regressor which implements the Openstf regressor API."""

    @staticmethod
    def _get_importance_names():
        return {
            "gain_importance_name": "total_gain",
            "weight_importance_name": "weight",
        }

    def fit(self, x, y, **kwargs):
        """Fit model."""
        super().fit(x, y)
        self.feature_importances_ = self._get_feature_importance_from_linear()
        return self

    def _get_feature_importance_from_linear(self):
        check_is_fitted(self)
        feature_importance_linear = np.abs(self.regressor_.coef_)
        reg_feature_importances_dict = dict(
            zip(self.non_null_columns_, feature_importance_linear)
        )
        return np.array(
            [reg_feature_importances_dict.get(c, 0) for c in self.feature_in_names_]
        )

    @property
    def feature_names(self):
        """The names of he features used to train the model."""
        return self.feature_in_names_

    @property
    def can_predict_quantiles(self):
        """Indicates wether this model can make quantile predictions."""
        return False



# SPDX-FileCopyrightText: 2017-2023 Contributors to the OpenSTEF project <korte.termijn.prognoses@alliander.com> # noqa E501>
#
# SPDX-License-Identifier: MPL-2.0
import logging
import os
from typing import Optional, Union, Tuple

import pandas as pd
import structlog

from openstef.data_classes.model_specifications import ModelSpecificationDataClass
from openstef.data_classes.prediction_job import PredictionJobDataClass
from openstef.exceptions import (
    InputDataInsufficientError,
    InputDataWrongColumnOrderError,
    OldModelHigherScoreError,
    SkipSaveTrainingForecasts,
)
from openstef.feature_engineering.feature_applicator import TrainFeatureApplicator
from openstef.metrics.reporter import Report, Reporter
from openstef.model.model_creator import ModelCreator
from openstef.model.regressors.regressor import OpenstfRegressor
from openstef.model.serializer import MLflowSerializer
from openstef.model.standard_deviation_generator import StandardDeviationGenerator
from openstef.model_selection.model_selection import split_data_train_validation_test
from openstef.validation import validation

DEFAULT_TRAIN_HORIZONS: list[float] = [0.25, 47.0]
MAXIMUM_MODEL_AGE: int = 7

DEFAULT_EARLY_STOPPING_ROUNDS: int = 10
PENALTY_FACTOR_OLD_MODEL: float = 1.2

logger = structlog.get_logger(__name__)


def train_model_pipeline(
    pj: PredictionJobDataClass,
    input_data: pd.DataFrame,
    check_old_model_age: bool,
    mlflow_tracking_uri: str,
    artifact_folder: str,
) -> Optional[tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]]:
    """Middle level pipeline that takes care of all persistent storage dependencies.

    Expected prediction jobs keys: "id",
    "model", "hyper_params", "feature_names".

    Args:
        pj: Prediction job
        input_data: Raw training input data
        check_old_model_age: Check if training should be skipped because the model is too young
        mlflow_tracking_uri: Tracking URI for MLFlow
        artifact_folder: Path where artifacts, such as trained models, are stored

    Returns:
        If pj.save_train_forecasts is False, None is returned
        Otherwise:
            - The train dataset with forecasts
            - The validation dataset with forecasts
            - The test dataset with forecasts

    """
    # Initialize logger and serializer
    serializer = MLflowSerializer(mlflow_tracking_uri=mlflow_tracking_uri)

    # Get old model and age
    old_model, model_specs, old_model_age = train_pipeline_step_load_model(
        pj, serializer
    )

    # Check old model age and continue yes/no
    if (old_model_age < MAXIMUM_MODEL_AGE) and check_old_model_age:
        logger.warning(
            f"Old model is younger than {MAXIMUM_MODEL_AGE} days, skip training"
        )
        if pj.save_train_forecasts:
            raise SkipSaveTrainingForecasts
        return

    # Train model with core pipeline
    try:
        model, report, model_specs_updated, data_sets = train_model_pipeline_core(
            pj,
            model_specs,
            input_data,
            old_model,
            horizons=pj.train_horizons_minutes,
        )
    except OldModelHigherScoreError as OMHSE:
        logger.error("Old model is better than new model", pid=pj["id"], exc_info=OMHSE)
        if pj.save_train_forecasts:
            raise SkipSaveTrainingForecasts from OMHSE
        return

    except InputDataInsufficientError as IDIE:
        logger.error(
            "Input data is insufficient after validation and cleaning",
            pid=pj["id"],
            exc_info=IDIE,
        )
        raise InputDataInsufficientError(IDIE)

    except InputDataWrongColumnOrderError as IDWCOE:
        logger.error(
            "Wrong column order, 'load' column should be first and 'horizon' column"
            " last.",
            pid=pj["id"],
            exc_info=IDWCOE,
        )
        raise InputDataWrongColumnOrderError(IDWCOE)

    # Save model and report. Report is always saved to MLFlow and optionally to disk
    serializer.save_model(
        model=model,
        experiment_name=str(pj["id"]),
        model_type=pj["model"],
        model_specs=model_specs_updated,
        report=report,
    )
    if artifact_folder:
        report_folder = os.path.join(artifact_folder, str(pj["id"]))
        Reporter.write_report_to_disk(report=report, report_folder=report_folder)

    # Clean up older models
    serializer.remove_old_models(
        experiment_name=str(pj["id"]), artifact_folder=artifact_folder
    )

    if pj.save_train_forecasts:
        return data_sets


def train_model_pipeline_core(
    pj: PredictionJobDataClass,
    model_specs: ModelSpecificationDataClass,
    input_data: pd.DataFrame,
    old_model: OpenstfRegressor = None,
    horizons: Union[list[float], str] = None,
) -> Union[
    OpenstfRegressor,
    Report,
    ModelSpecificationDataClass,
    tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame],
]:
    """Train model core pipeline.

    Trains a new model given a prediction job, input data and compares it to an old model.
    This pipeline has no database or persistent storage dependencies.

    Args:
        pj: Prediction job
        model_specs: Dataclass containing model specifications
        input_data: Input data
        old_model: Old model to compare to. Defaults to None.
        horizons: horizons to train on in hours.

    Raises:
        InputDataInsufficientError: when input data is insufficient.
        InputDataWrongColumnOrderError: when input data has a invalid column order.
        OldModelHigherScoreError: When old model is better than new model.

    Returns:
        - Fitted_model (OpenstfRegressor)
        - Report (Report)
        - Modelspecs (ModelSpecificationDataClass)
        - Datasets (tuple[pd.DataFrmae, pd.DataFrame, pd.Dataframe): The train, validation and test sets

    """
    if horizons is None:
        if pj.train_horizons_minutes is None:
            horizons = DEFAULT_TRAIN_HORIZONS
        else:
            horizons = pj.train_horizons_minutes

    logger = structlog.get_logger(__name__)

    # Call common pipeline
    model, report, train_data, validation_data, test_data = train_pipeline_common(
        pj,
        model_specs,
        input_data,
        horizons,
    )
    model_specs.feature_names = list(train_data.columns)

    # Check if new model is better than old model
    if old_model:
        combined = pd.concat([train_data, validation_data])
        # skip the forecast column added at the end of dataframes
        if pj.save_train_forecasts:
            combined = combined.iloc[:, :-1]

        x_data, y_data = (
            combined.iloc[:, 1:-1],
            combined.iloc[:, 0],
        )

        # Score method always returns R^2
        score_new_model = model.score(x_data, y_data)

        # Try to compare new model to old model.
        # If this does not success, for example since the feature names of the
        # old model differ from the new model, the new model is considered better
        try:
            score_old_model = old_model.score(x_data, y_data)

            # Check if R^2 is better for old model
            if score_old_model > score_new_model * PENALTY_FACTOR_OLD_MODEL:
                raise OldModelHigherScoreError(
                    f"Old model is better than new model for {pj['id']}."
                )

            logger.info(
                "New model is better than old model, continuing with training procces"
            )
        except ValueError as e:
            logger.info("Could not compare to old model", pid=pj["id"], exc_info=e)

    return model, report, model_specs, (train_data, validation_data, test_data)


def train_pipeline_common(
    pj: PredictionJobDataClass,
    model_specs: ModelSpecificationDataClass,
    input_data: pd.DataFrame,
    horizons: Union[list[float], str],
    test_fraction: float = 0.0,
    backtest: bool = False,
    test_data_predefined: pd.DataFrame = pd.DataFrame(),
) -> tuple[OpenstfRegressor, Report, pd.DataFrame, pd.DataFrame, pd.DataFrame]:
    """Common pipeline shared with operational training and backtest training.

    Args:
        pj: Prediction job
        model_specs: Dataclass containing model specifications
        input_data: Input data
        horizons: horizons to train on in hours.
        test_fraction: fraction of data to use for testing
        backtest: boolean if we need to do a backtest
        test_data_predefined: Predefined test data frame to be used in the pipeline
            (empty data frame by default)

    Returns:
        - The trained model
        - Report
        - The train data
        - The validation data
        - The test data

    Raises:
        InputDataInsufficientError: when input data is insufficient.
        InputDataWrongColumnOrderError: when input data has a invalid column order.

    """
    data_with_features = train_pipeline_step_compute_features(
        pj=pj,
        model_specs=model_specs,
        input_data=input_data,
        horizons=horizons,
    )

    train_data, validation_data, test_data = train_pipeline_step_split_data(
        data_with_features=data_with_features,
        pj=pj,
        test_fraction=test_fraction,
        backtest=backtest,
        test_data_predefined=test_data_predefined,
    )

    model = train_pipeline_step_train_model(
        pj=pj,
        model_specs=model_specs,
        train_data=train_data,
        validation_data=validation_data,
    )

    # Report about the training process
    reporter = Reporter(train_data, validation_data, test_data)
    report = reporter.generate_report(model)

    if pj.save_train_forecasts:
        train_data["forecast"] = model.predict(train_data.iloc[:, 1:-1])
        validation_data["forecast"] = model.predict(validation_data.iloc[:, 1:-1])
        test_data["forecast"] = model.predict(test_data.iloc[:, 1:-1])

    return model, report, train_data, validation_data, test_data


def train_pipeline_step_load_model(
    pj: PredictionJobDataClass, serializer: MLflowSerializer
) -> tuple[OpenstfRegressor, ModelSpecificationDataClass, Union[int, float]]:
    try:
        old_model, model_specs = serializer.load_model(experiment_name=str(pj.id))
        old_model_age = old_model.age  # Age attribute is openstef specific
        return old_model, model_specs, old_model_age
    except (AttributeError, FileNotFoundError, LookupError):
        logger.warning("No old model found, training new model", pid=pj.id)
    except Exception:
        logger.exception("Old model could not be loaded, training new model", pid=pj.id)
    old_model = None
    old_model_age = float("inf")
    if pj["default_modelspecs"] is not None:
        model_specs = pj["default_modelspecs"]
        if model_specs.id != pj.id:
            raise RuntimeError(
                "The id of the prediction job and its default model_specs do not"
                " match."
            )
    else:
        # create basic model_specs
        model_specs = ModelSpecificationDataClass(id=pj["id"])

    return old_model, model_specs, old_model_age


def train_pipeline_step_compute_features(
    pj: PredictionJobDataClass,
    model_specs: ModelSpecificationDataClass,
    input_data: pd.DataFrame,
    horizons=list[float],
) -> pd.DataFrame:
    """Compute features and perform consistency checks.

    Args:
        pj: Prediction job
        model_specs: Dataclass containing model specifications
        input_data: Input data
        horizons: horizons to train on in hours.

    Returns:
        The dataframe with features need to train the model

    Raises:
        InputDataInsufficientError: when input data is insufficient.
        InputDataWrongColumnOrderError: when input data has a invalid column order.
        ValueError: when the horizon is a string and the corresponding column in not in the input data

    """
    if pj["model"] == "proloaf":
        # proloaf is only able to train with one horizon
        horizons = [horizons[0]]

    if input_data.empty:
        raise InputDataInsufficientError("Input dataframe is empty")
    elif "load" not in input_data.columns:
        raise InputDataWrongColumnOrderError(
            "Missing the load column in the input dataframe"
        )

    if isinstance(horizons, str):
        if horizons not in set(input_data.columns):
            raise ValueError(
                f"The horizon parameter specifies a column name ({horizons}) missing in"
                " the input data."
            )
        else:
            # sort data to avoid same date repeated multiple time
            input_data = input_data.sort_values(horizons)
    # Validate and clean data
    validated_data = validation.drop_target_na(
        validation.validate(pj["id"], input_data, pj["flatliner_treshold"])
    )
    # Check if sufficient data is left after cleaning
    if not validation.is_data_sufficient(
        validated_data,
        pj["completeness_treshold"],
        pj["minimal_table_length"],
    ):
        raise InputDataInsufficientError(
            "Input data is insufficient, after validation and cleaning"
        )

    # Custom data prep or legacy behavior
    if pj.data_prep_class:
        data_prep_class, data_prep_args = pj.data_prep_class.load()
        data_with_features = data_prep_class(
            pj=pj,
            model_specs=model_specs,
            horizons=horizons,
            **data_prep_args,
        ).prepare_train_data(validated_data)
    else:
        data_with_features = TrainFeatureApplicator(
            horizons=horizons,
            feature_names=model_specs.feature_names,
            feature_modules=model_specs.feature_modules,
        ).add_features(validated_data, pj=pj)

    return data_with_features


def train_pipeline_step_train_model(
    pj: PredictionJobDataClass,
    model_specs: ModelSpecificationDataClass,
    train_data: pd.DataFrame,
    validation_data: pd.DataFrame,
) -> OpenstfRegressor:
    """Train the model.

    Args:
        pj: Prediction job
        model_specs: Dataclass containing model specifications
        train_data: The training data
        validation_data: The test data

    Returns:
        The trained model

    """
    # Test if first column is "load" and last column is "horizon"
    if train_data.columns[0] != "load" or train_data.columns[-1] != "horizon":
        raise InputDataWrongColumnOrderError(
            f"Wrong column order for {pj['id']} "
            "'load' column should be first and 'horizon' column last."
        )

    # Create relevant model
    model = ModelCreator.create_model(
        pj["model"],
        quantiles=pj["quantiles"],
    )

    # split x and y data
    train_x, train_y = train_data.iloc[:, 1:-1], train_data.iloc[:, 0]
    validation_x, validation_y = (
        validation_data.iloc[:, 1:-1],
        validation_data.iloc[:, 0],
    )

    # Configure evals for early stopping
    eval_set = [(train_x, train_y), (validation_x, validation_y)]

    # Set relevant hyperparameters
    # define protected hyperparams which are derived from prediction_job
    protected_hyperparams = ["quantiles"]
    valid_hyper_parameters = {
        key: value
        for key, value in model_specs.hyper_params.items()
        if key in model.get_params().keys() and key not in protected_hyperparams
    }

    # Add early stopping to set_params if this is supported by the model
    if "early_stopping_rounds" in model.get_params().keys():
        valid_hyper_parameters.update(
            dict(early_stopping_rounds=DEFAULT_EARLY_STOPPING_ROUNDS)
        )

    # Temporary fix to allow xgboost version upgrade -> set n_estimators if present and None
    if not valid_hyper_parameters.get("n_estimators", True):
        valid_hyper_parameters.update(dict(n_estimators=100))
        logging.info("Deprecation warning: n_estimators=None found, overwriting.")

    model.set_params(**valid_hyper_parameters)
    model.fit(
        train_x,
        train_y,
        eval_set=eval_set,
        verbose=False,
    )
    # Gets the feature importance df or None if we don't have feature importance
    model.feature_importance_dataframe = model.set_feature_importance()

    logging.info("Fitted a new model, not yet stored")

    # Do confidence interval determination
    model = StandardDeviationGenerator(
        validation_data
    ).generate_standard_deviation_data(model)

    return model


def train_pipeline_step_split_data(
    data_with_features: pd.DataFrame,
    pj: PredictionJobDataClass,
    test_fraction: float,
    backtest: bool = False,
    test_data_predefined: pd.DataFrame = pd.DataFrame(),
) -> Union[pd.DataFrame, pd.DataFrame, pd.DataFrame]:
    """The default way to perform train, val, test split.

    Args:
        data_with_features: Input data
        pj: Prediction job
        test_fraction: fraction of data to use for testing
        backtest: boolean if we need to do a backtest
        test_data_predefined: Predefined test data frame to be used in the pipeline
            (empty data frame by default)

    Returns:
        - Train dataset
        - Validation dataset
        - Test dataset

    """
    # if test_data is predefined, apply the pipeline only on the remaining data
    if not test_data_predefined.empty:
        test_data_predefined = data_with_features[
            data_with_features.index.isin(test_data_predefined.index)
        ].sort_index()
        data_with_features = data_with_features[
            ~data_with_features.index.isin(test_data_predefined.index)
        ].sort_index()

    # Split data
    if pj.train_split_func is None:
        split_func = split_data_train_validation_test
        split_args = {
            "stratification_min_max": pj["model"] != "proloaf",
            "back_test": backtest,
        }
    else:
        split_func, split_args = pj.train_split_func.load(
            required_arguments=["data", "test_fraction"]
        )

    train_data, validation_data, test_data = split_func(
        data_with_features, test_fraction, **split_args
    )

    # if test_data is predefined, use this over the returned test_data of split function
    if not test_data_predefined.empty:
        test_data = test_data_predefined

    return train_data, validation_data, test_data
